import os
import unittest
import six.moves.urllib.parse as urllibparse

from datarobot.helpers.partitioning_methods import BasePartitioningMethod, GroupTVH


_here = os.path.dirname(os.path.abspath(__file__))
FIXTURES_DIR = os.path.join(_here, 'fixtures')


def fixture_file_path(filename):
    return os.path.join(FIXTURES_DIR, filename)


class URLParamsTestCase(unittest.TestCase):

    def assert_has_params(self, url, **params):
        query_string = urllibparse.unquote(url).split('?')[1]
        query_parts = query_string.split('&')
        query_pairs = [param.split('=') for param in query_parts]
        query_params = {pair[0]: pair[1]
                        for pair in query_pairs}

        for key, value in query_params.items():
            self.assertIn(key, query_params,
                          'Expected parameter {} not found'.format(key))
            self.assertEqual(query_params[key], params[key])


class HelpersTestCase(unittest.TestCase):

    def test_cant_create_without_cv(self):
        with self.assertRaises(TypeError):
            BasePartitioningMethod(validation_type='CV')

        with self.assertRaises(TypeError):
            BasePartitioningMethod(cv_method='user')

    def test_partitioning_repr(self):
        p_method = GroupTVH(holdout_pct=1, validation_pct=2,
                            partition_key_cols=['one_feature', 'sec feature'],
                            seed=42)
        assert repr(p_method).startswith('GroupTVH({')

    def test_create_with_right(self):
        p_method = BasePartitioningMethod(validation_type='CV',
                                          cv_method='user')
        self.assertEqual(p_method.validation_type, 'CV')
        self.assertEqual(p_method.cv_method, 'user')

    def test_collect_payload(self):
        p_method = BasePartitioningMethod(validation_type='CV',
                                          cv_method='user')
        p = p_method.collect_payload()
        self.assertTrue('validation_type' in p)
        self.assertTrue('cv_method' in p)
        self.assertTrue('partition_key_cols' in p)
        self.assertEqual(p['validation_type'], 'CV')
        self.assertIsNone(p['partition_key_cols'])



import unittest
import contextlib
import warnings
import mock

from datarobot.client import set_client
from datarobot.rest import RESTClientObject


class SDKTestcase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        set_client(None)
        cls.patcher = mock.patch('datarobot.client._get_endpoint_from_file')
        c = RESTClientObject(auth='t-token', endpoint='https://host_name.com')
        cls.patcher.return_value = None
        cls.patcher.start()
        set_client(c)

    @classmethod
    def tearDownClass(cls):
        cls.patcher.stop()
        set_client(None)


def assertJsonEq(first, second, msg='Ouups'):
    assert sorted(first) == sorted(second), '%r != %r\n%s' % (first,
                                                              second,
                                                              msg)


@contextlib.contextmanager
def warns(*categories):
    """Assert warnings are raised in a block, analogous to pytest.raises."""
    with warnings.catch_warnings(record=True) as messages:
        yield messages
    assert tuple(message.category for message in messages) == categories



# flake8: noqa

__version__ = '2.2.32'

from .enums import (
    SCORING_TYPE,
    QUEUE_STATUS,
    AUTOPILOT_MODE,
    VERBOSITY_LEVEL,
)
from .client import Client
from .errors import AppPlatformError
from .helpers import *
from .models import (
    Project,
    Model,
    ModelJob,
    Blueprint,
    Featurelist,
    Feature,
    PredictJob,
    Job
)




import six
import trafaret as t

from .. import errors
from ..client import get_client, staticproperty
from ..enums import JOB_TYPE, MODEL_JOB_STATUS
from .job import Job
from ..utils import from_api, retry


def wait_for_async_model_creation(project_id, model_job_id, max_wait=600):
    """
    Given a Project id and ModelJob id poll for status of process
    responsible for model creation until model is created.

    Parameters
    ----------
    project_id : str
        The identifier of the project

    model_job_id : str
        The identifier of the ModelJob

    max_wait : int, optional
        Time in seconds after which model creation is considered
        unsuccessful

    Returns
    -------
    model : Model
        Newly created model

    Raises
    ------
    AsyncModelCreationError
        Raised if status of fetched ModelJob object is ``error``
    AsyncTimeoutError
        Model wasn't created in time, specified by ``max_wait`` parameter
    """
    for _ in retry.wait(max_wait):
        try:
            model_job = ModelJob.get(project_id, model_job_id)
        except errors.PendingJobFinished:
            return ModelJob.get_model(project_id, model_job_id)
        if model_job.status == MODEL_JOB_STATUS.ERROR:
            e_msg = 'Model creation unsuccessful'
            raise errors.AsyncModelCreationError(e_msg)

    timeout_msg = 'Model creation timed out in {} seconds'.format(max_wait)
    raise errors.AsyncTimeoutError(timeout_msg)


class ModelJob(Job):
    _client = staticproperty(get_client)
    _fields = frozenset(['id',
                         'status',
                         'sample_pct',
                         'model_type',
                         'processes'])

    _path = 'projects/{}/modelJobs/{}/'

    def __repr__(self):
        return 'ModelJob({}, status={})'.format(self.model_type,
                                                self.status)  # pragma:no cover

    def __init__(self, data):
        self.id = None
        self.blueprint = None
        self.project = None
        self.processes = None
        self.job_type = JOB_TYPE.MODEL
        self.model_type = None
        self.status = None
        self.sample_pct = None
        self.featurelist_id = None

        self._data = data
        if isinstance(data, dict):
            self._set_up(data)
        else:
            raise ValueError

    def _set_up(self, data):
        from . import Project, Blueprint

        data = from_api(data)
        # Construction Project
        self.project = Project(data.get('project_id'))

        # Construction Blueprint
        bp_data = {'id': data.get('blueprint_id'),
                   'processes': data.get('processes'),
                   'model_type': data.get('model_type'),
                   'project_id': data.get('project_id')}
        self.blueprint = Blueprint(bp_data)

        converter = t.Dict({
            t.Key('id', optional=True) >> 'id': t.Int,
            t.Key('status', optional=True) >> 'status': t.String,
            t.Key('sample_pct', optional=True) >> 'sample_pct': t.Float,
            t.Key('model_type', optional=True) >> 'model_type': t.String,
            t.Key('processes', optional=True) >> 'processes': t.List(t.String),
            t.Key('featurelistId', optional=True) >> 'featurelist_id': t.List(t.String),
        }).allow_extra("*")

        for k, v in six.iteritems(converter.check(data)):
            if k in self._fields:
                setattr(self, k, v)

        self._path = self._path.format(data.get('project_id'), data.get('id'))

    @classmethod
    def from_job(cls, job):
        """ Transforms a generic Job into a ModelJob

        Parameters
        ----------
        job: Job
            A generic job representing a ModelJob

        Returns
        -------
        model_job: ModelJob
            A fully populated ModelJob with all the details of the job

        Raises
        ------
        ValueError:
            If the generic Job was not a model job, e.g. job_type != JOB_TYPE.MODEL
        """
        if not job.job_type == JOB_TYPE.MODEL:
            raise ValueError('wrong job_type: {}'.format(job.job_type))
        data = job._get_full_details()
        return ModelJob(data)

    @classmethod
    def get(cls, project_id, model_job_id):
        """
        Fetches one ModelJob. If the job finished, raises PendingJobFinished
        exception.

        Parameters
        ----------
        project_id : str
            The identifier of the project the model belongs to
        model_job_id : str
            The identifier of the model_job

        Returns
        -------
        model_job : ModelJob
            The pending ModelJob

        Raises
        ------
        PendingJobFinished
            If the job being queried already finished, and the server is
            re-routing to the finished model.
        AsyncFailureError
            Querying this resource gave a status code other than 200 or 303
        """
        url = cls._path.format(project_id, model_job_id)
        response = cls._client.get(url, allow_redirects=False)
        if response.status_code == 200:
            data = response.json()
            return cls(data)
        elif response.status_code == 303:
            raise errors.PendingJobFinished
        else:
            e_msg = 'Server unexpectedly returned status code {}'
            raise errors.AsyncFailureError(
                e_msg.format(response.status_code))

    @classmethod
    def get_model(cls, project_id, model_job_id):
        """
        Fetches a finished model from the job used to create it.

        Parameters
        ----------
        project_id : str
            The identifier of the project the model belongs to
        model_job_id : str
            The identifier of the model_job

        Returns
        -------
        model : Model
            The finished model

        Raises
        ------
        JobNotFinished
            If the job has not finished yet
        AsyncFailureError
            Querying the model_job in question gave a status code other than 200 or
            303
        """
        from . import Model

        url = cls._path.format(project_id, model_job_id)
        response = cls._client.get(url, allow_redirects=False)
        if response.status_code == 200:
            data = response.json()
            raise errors.JobNotFinished('Pending job status is {}'.format(data['status']))
        elif response.status_code == 303:
            location = response.headers['Location']
            model_data = Model.fetch_resource_data(location, join_endpoint=False)
            return Model(data=model_data)
        else:
            e_msg = 'Server unexpectedly returned status code {}'
            raise errors.AsyncFailureError(
                e_msg.format(response.status_code))

    def cancel(self):
        """
        Cancel this job. If this job has not finished running, it will be
        removed and canceled.
        """
        self._client.delete(self._path)



import pytest
from .utils import SDKTestcase
from datarobot.client import get_client


@pytest.yield_fixture(scope='module')
def client():
    """A mocked client."""
    SDKTestcase.setUpClass()
    yield get_client()
    SDKTestcase.tearDownClass()



from datarobot import Featurelist, Project
from .utils import SDKTestcase

import responses


class TestFeaturelist(SDKTestcase):

    def test_instantiate_featurelist(self):
        data = {
            'id': '5223deadbeefdeadbeef9999',
            'name': 'Raw Features',
            'features': ['One Fish', 'Two Fish', 'Read Fish', 'Blue Fish'],
            'project_id': '5223deadbeefdeadbeef0101'
        }

        flist = Featurelist(data)

        self.assertEqual(flist.id, data['id'])
        self.assertEqual(flist.name, data['name'])
        self.assertEqual(flist.features, data['features'])
        self.assertEqual(flist.project.id, data['project_id'])
        self.assertEqual(repr(flist), 'Featurelist(Raw Features)')

    def test_value_erro_with_bad_constructor_data(self):
        data = 'Green eggs'
        with self.assertRaises(ValueError):
            Featurelist(data)


class TestGet(SDKTestcase):

    def setUp(self):
        self.raw_return = """
        {
        "id": "f-id",
        "project_id": "p-id",
        "name": "My Feature List",
        "features": ["One Fish", "Two Fish", "Red Fish", "Blue Fish"]
        }
        """

    @responses.activate
    def test_get_featurelist(self):
        """
        Test get project
        """
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/featurelists/f-id/',
                      body=self.raw_return,
                      status=200,
                      content_type='application/json')
        result = Featurelist.get('p-id', 'f-id')
        self.assertEqual(result.project.id, 'p-id')
        self.assertEqual(result.name, 'My Feature List')

    @responses.activate
    def test_get_featurelist_with_project_instance(self):
        responses.add(responses.GET,
                      'https://host_name.com/projects/p-id/featurelists/f-id/',
                      body=self.raw_return,
                      status=200,
                      content_type='application/json')

        pdata = {
            'id': 'p-id',
            'project_name': 'Projects'
        }
        project = Project(pdata)

        result = Featurelist.get(project, 'f-id')
        self.assertEqual(result.project.id, 'p-id')
        self.assertEqual(result.name, 'My Feature List')

    def test_rejects_bad_project_input(self):
        not_a_project = 5
        with self.assertRaises(ValueError):
            Featurelist.get(not_a_project, 'f-id')



import logging
import os
from six.moves import configparser

import yaml
from .utils import deprecated
from .rest import RESTClientObject

logger = logging.getLogger(__package__)

__all__ = ('Client', 'get_client', 'set_client')

_global_client = None


def Client(username=None, password=None, token=None, endpoint=None, config_path=None):
    """Return global `RESTClientObject` with optional configuration.
    Missing configuration will be read from env or config file.

    Parameters
    ----------
    token : str, optional
        API token
    endpoint : str, optional
        Base url of API
    config_path : str, optional
        Alternate location of config file
    """
    global _global_client
    no_username_password = ('Username and password authentication is no longer'
                            ' supported - please use an API token')
    if username or password:
        raise ValueError(no_username_password)
    if config_path:
        if not _file_exists(config_path):
            raise ValueError('Invalid config path - no file at {}'.format(config_path))
    if not token:
        if config_path:
            username, password, token = _get_credentials_from_file(config_path)
        else:
            username, password, token = get_credentials_from_out()
        if not (username and password) and not token:
            raise ValueError('Credentials were improperly configured - could not find token')
    if not endpoint:
        if config_path:
            endpoint = _get_endpoint_from_file(config_path)
        else:
            endpoint = _get_endpoint()
    if username and password:
        raise ValueError(no_username_password)
    else:
        _global_client = RESTClientObject(auth=token, endpoint=endpoint)
    return _global_client


def get_client():
    return _global_client or Client()


class staticproperty(property):
    def __get__(self, instance, owner):
        return self.fget()


def set_client(client):
    """
    Set the global HTTP client for sdk.
    Returns previous client.
    """
    global _global_client
    previous = _global_client
    _global_client = client
    return previous


def get_credentials_from_out():
    username, password, token = [None] * 3
    if 'DATAROBOT_USERNAME' in os.environ:
        username = os.environ.get('DATAROBOT_USERNAME')
        password = os.environ.get('DATAROBOT_PASSWORD')
    elif 'DATAROBOT_API_TOKEN' in os.environ:
        token = os.environ.get('DATAROBOT_API_TOKEN')
    elif _get_config_file():
        config_file = _get_config_file()
        username, password, token = _get_credentials_from_file(config_file)
    return username, password, token


def _get_credentials_from_yaml(config_path):
    with open(config_path, "r") as config_file:
        config = yaml.safe_load(config_file)
    username = config.get("username")
    password = config.get("password")
    token = config.get("token")
    return username, password, token


def _get_credentials_from_ini(config_path):
    cfg_parser = configparser.RawConfigParser()
    cfg_parser.read(config_path)
    username, password, token = None, None, None
    if cfg_parser.has_option('datarobot', 'username'):
        username = cfg_parser.get('datarobot', 'username')
    if cfg_parser.has_option('datarobot', 'password'):
        password = cfg_parser.get('datarobot', 'password')
    if cfg_parser.has_option('datarobot', 'token'):
        token = cfg_parser.get('datarobot', 'token')
    return username, password, token


def path_is_yaml(path):
    _, file_extension = os.path.splitext(path)
    return file_extension == '.yaml'


def _get_credentials_from_file(config_path):
    if path_is_yaml(config_path):
        return _get_credentials_from_yaml(config_path)
    else:
        # put deprecation warning here
        # @deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3',
        #     message='Use a YAML configuration file (default location: '
        #             '~/.config/datarobot/drconfig.yaml) instead.')
        return _get_credentials_from_ini(config_path)


def _get_endpoint_from_yaml(config_path):
    with open(config_path, "r") as config_file:
        config = yaml.safe_load(config_file)
        return config.get('endpoint')


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3',
            message='Use a YAML configuration file (default location: '
                    '~/.config/datarobot/drconfig.yaml) instead.')
def _get_endpoint_from_ini(config_path):
    if config_path:
        cfg_parser = configparser.RawConfigParser()
        cfg_parser.read(config_path)
        try:
            return cfg_parser.get('datarobot', 'endpoint')
        except (configparser.NoSectionError,
                configparser.NoOptionError):
            return None
    return None


def _get_endpoint_from_file(config_path):
    if config_path is None:
        return None
    elif path_is_yaml(config_path):
        return _get_endpoint_from_yaml(config_path)
    else:
        return _get_endpoint_from_ini(config_path)


def _get_endpoint():
    endpoint = os.environ.get('DATAROBOT_ENDPOINT')
    if endpoint is not None:
        return endpoint
    endpoint = _get_endpoint_from_file(_get_config_file())
    if endpoint is not None:
        return endpoint
    raise ValueError('Improper Configuration - no endpoint found')


def _get_config_file():
    first_choice_config_path = os.path.expanduser('~/.config/datarobot/drconfig.yaml')
    deprecated_config_path = os.path.join(os.path.expanduser('~'), '.datarobotrc')
    if 'DATAROBOT_CONFIG_FILE' in os.environ:
        config_path = os.environ.get('DATAROBOT_CONFIG_FILE')
        if _file_exists(config_path):
            return config_path
        else:
            raise ValueError('Environment variable DATAROBOT_CONFIG_FILE points to an invalid '
                             'config file')
    elif _file_exists(first_choice_config_path):
        return first_choice_config_path
    elif _file_exists(deprecated_config_path):
        return deprecated_config_path
    else:
        return None


_file_exists = os.path.isfile



import webbrowser

import six
import trafaret as t

from ..client import get_client, staticproperty
from ..utils import from_api, get_id_from_response


class Model(object):
    _client = staticproperty(get_client)
    _path = 'projects/{}/models/'
    _fields = frozenset(['id', 'project', 'processes', 'blueprint_id',
                         'model_type', 'model_category', 'featurelist_id',
                         'sample_pct', 'metrics'])
    _converter = t.Dict({
        t.Key('id', optional=True): t.String,
        t.Key('processes', optional=True): t.List(t.Or(t.String, t.Null)),
        t.Key('featurelist_name', optional=True): t.String,
        t.Key('featurelist_id', optional=True): t.String,
        t.Key('project_id', optional=True): t.String,
        t.Key('sample_pct', optional=True): t.Float,
        t.Key('model_type', optional=True): t.String,
        t.Key('model_category', optional=True): t.String,
        t.Key('blueprint_id', optional=True): t.String,
        t.Key('metrics', optional=True): t.Dict().allow_extra("*"),
    })

    def __init__(self, data, project=None):
        self._data = data
        self.project = project
        self.blueprint = None
        self.featurelist = None
        if isinstance(data, dict):
            self._set_up(data)
        elif isinstance(data, tuple):
            self.id = data[1]
            from . import Project
            self.project = Project(data[0])
        else:
            raise ValueError
        self.project_id = self.project.id
        self._fix_path()

    def __repr__(self):
        return 'Model({})'.format(self.model_type or self.id)  # pragma:no cover

    def _fix_path(self):
        self._path = self._path.format(self.project.id)

    def _set_up(self, data):
        from . import Project, Blueprint, Featurelist
        data = self._converter.check(from_api(data, do_recursive=False))
        # Construction Project
        if not self.project:
            self.project = Project(data.get('project_id'))

        # Construction Blueprint
        bp_data = {'id': data.get('blueprint_id'),
                   'processes': data.get('processes'),
                   'model_type': data.get('model_type'),
                   'project_id': data.get('project_id')}
        self.blueprint = Blueprint(bp_data)

        # Construction FeatureList
        ft_list_data = {'id': data.get('featurelist_id'),
                        'name': data.get('featurelist_name')}
        self.featurelist = Featurelist(ft_list_data)

        for k, v in six.iteritems(data):
            if k in self._fields:
                setattr(self, k, v)

    @classmethod
    def get(cls, project, model_id):
        """
        Retrieve a specific model.

        Parameters
        ----------
        project : Project instance or str.
            You can provide a project instance or project id to identify
            which project the model is associated with.
        model_id : str
            The ``model_id`` of the leaderboard item to retrieve.

        Returns
        -------
        model : Model
            The queried instance.

        Raises
        ------
        ValueError
            passed ``project`` parameter value is of not supported type
        """
        from . import Project
        project_instance = False
        if isinstance(project, Project):
            project_id = project.id
            project_instance = True
        elif isinstance(project, six.string_types):
            project_id = project
        else:
            raise ValueError('Project arg can be Project instance or str')
        url = cls._path.format(project_id) + model_id + '/'
        resp_data = cls.fetch_resource_data(url)
        return cls(resp_data, project=project if project_instance else None)

    @classmethod
    def fetch_resource_data(cls, url, join_endpoint=True):
        """
        Used to acquire model data directly from its url.

        Consider using `get` instead, as this is a convenience function
        used for development of datarobot

        Parameters
        ----------
        url : string
            The resource we are acquiring
        join_endpoint : boolean, optional
            Whether the client's endpoint should be joined to the URL before
            sending the request. Location headers are returned as absolute
            locations, so will _not_ need the endpoint

        Returns
        -------
        model_data : dict
            The queried model's data
        """
        return cls._client.get(url, join_endpoint=join_endpoint).json()

    def get_features_used(self):
        """Query the server to determine which features were used.

        Note that the data returned by this method is possibly different
        than the names of the features in the featurelist used by this model.
        This method will return the raw features that must be supplied in order
        for predictions to be generated on a new set of data. The featurelist,
        in contrast, would also include the names of derived features.

        Returns
        -------
        features : list of str
            The names of the features used in the model.
        """
        url = '{}{}/features/'.format(self._path, self.id)
        resp_data = self._client.get(url).json()
        return resp_data['featureNames']

    def delete(self):
        """
        Delete a model from the project's leaderboard.
        """
        url = '{}{}/'.format(self._path, self.id)
        self._client.delete(url)

    def get_leaderboard_ui_permalink(self):
        """
        Returns
        -------
        url : str
            Permanent static hyperlink to this model at leaderboard.
        """
        return '{}/{}{}'.format(self._client.domain, self._path, self.id)

    def open_model_browser(self):
        """
        Opens model at project leaderboard in web browser.

        Note:
        If text-mode browsers are used, the calling process will block
        until the user exits the browser.
        """

        url = self.get_leaderboard_ui_permalink()
        return webbrowser.open(url)

    def train(self, sample_pct=None, featurelist_id=None, scoring_type=None):
        """
        Train this model on `sample_pct` percents.
        This method creates a new training job for worker and appends it to
        the end of the queue for this project.
        After the job has finished you can get this model by retrieving
        it from the project leaderboard.

        Parameters
        ----------
        sample_pct : float, optional
            The amount of data to use for training. Defaults to the maximum
            amount available based on the project configuration.
        featurelist_id : str, optional
            The identifier of the featurelist to use. If not defined, the
            featurelist of this model is used.
        scoring_type : str, optional
            Either ``SCORING_TYPE.validation`` or
            ``SCORING_TYPE.cross_validation``. ``SCORING_TYPE.validation``
            is available for every partitioning type, and indicates that
            the default model validation should be used for the project.
            If the project uses a form of cross-validation partitioning,
            ``SCORING_TYPE.cross_validation`` can also be used to indicate
            that all of the available training/validation combinations
            should be used to evaluate the model.

        Returns
        -------
        model_job_id : str
            id of created job, can be used as parameter to ``ModelJob.get``
            method or ``wait_for_async_model_creation`` function

        Examples
        --------
        .. code-block:: python

            model = Model.get('p-id', 'l-id')
            model_job_id = model.train()
        """
        url = self._path
        payload = {
            'blueprint_id': self.blueprint_id,
        }
        if sample_pct is not None:
            payload['sample_pct'] = sample_pct
        if scoring_type is not None:
            payload['scoring_type'] = scoring_type
        if featurelist_id is not None:
            payload['featurelist_id'] = featurelist_id
        else:
            payload['featurelist_id'] = self.featurelist_id
        response = self._client.post(url, data=payload)
        return get_id_from_response(response)



import six
from ..utils import from_api


class Blueprint(object):
    _fields = frozenset(['id', 'processes', 'model_type',
                         'project_id'])

    def __init__(self, data):
        self.id = None
        self.processes = None
        self.model_type = None
        self.project_id = None

        self._data = data
        if isinstance(data, dict):
            self._set_up(from_api(data))
        else:
            raise ValueError

    def __repr__(self):
        return 'Blueprint({})'.format(self.model_type)  # pragma:no cover

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def _set_up(self, data):
        for k, v in six.iteritems(data):
            if k in self._fields:
                setattr(self, k, v)



import datetime
import unittest
import pandas as pd
import responses

from datarobot import errors
from datarobot.client import Client, set_client
from datarobot.errors import DataRobotDeprecationWarning
from datarobot.utils import (dataframe_to_buffer, parse_time, get_id_from_response,
                             from_api, deprecation, recognize_sourcedata, is_urlsource)
from .test_helpers import fixture_file_path
from .utils import warns


class TestDataframeSerialization(unittest.TestCase):

    def test_no_index_please(self):
        df = pd.DataFrame({'a': range(100), 'b': range(100)})
        buff = dataframe_to_buffer(df)
        readback = pd.read_csv(buff)
        self.assertEqual(readback.columns.tolist(), ['a', 'b'])

    def test_parse_time(self):
        self.assertEqual('BAD TIME', parse_time('BAD TIME'))  # returns value
        test_string_time = datetime.datetime.now().isoformat()
        self.assertIsInstance(test_string_time, str)
        self.assertIsInstance(parse_time(test_string_time), datetime.datetime)

    @responses.activate
    def test_get_id_from_response_location_header(self):
        responses.add(responses.POST,
                      'http://nothing/',
                      body='',
                      adding_headers={'Location': 'http://nothing/f-id/'})
        client = set_client(Client(token='no_matter',
                                   endpoint='http://nothing'))
        resp = client.post('')
        self.assertEqual(get_id_from_response(resp), 'f-id')


class TestFromAPI(unittest.TestCase):

    def test_nested_list_of_objects_all_changed(self):
        source = {
            'oneFish': [
                {'twoFish': 'redFish'},
                {'blueFish': 'noFish'}
            ]
        }
        result = from_api(source)
        inner = result['one_fish']
        self.assertEqual(inner[0]['two_fish'], 'redFish')
        self.assertEqual(inner[1]['blue_fish'], 'noFish')

    def test_nested_objects_all_changed(self):
        source = {
            'oneFish': {
                'twoFish': 'redFish'
            }
        }

        result = from_api(source)
        self.assertEqual(result['one_fish']['two_fish'], 'redFish')


@deprecation.deprecated(deprecated_since_version='v0.1.2',
                        will_remove_version='v1.2.3')
def bar(arg1, arg2):
    """
    A dummy function to use for testing the deprecation mechanisms

    Parameters
    ----------
    arg1 : int
    arg2 : int

    Returns
    -------
    total : int
        arg1 + arg2

    """
    return arg1 + arg2


@deprecation.deprecated(deprecated_since_version='v0.1.2',
                        will_remove_version='v1.2.3',
                        message='Use of `bar` is recommended instead.')
def foo(arg1, arg2):
    """
    A dummy function to use for testing the deprecation mechanisms

    Parameters
    ----------
    arg1 : int
    arg2 : int

    Returns
    -------
    total : int
        foo + bar

    """
    return arg1 + arg2


class TestDeprecation(unittest.TestCase):

    def test_deprecation_with_message(self):
        with warns(DataRobotDeprecationWarning) as messages:
            foo(1, 2)
        assert str(messages[-1].message) == \
            '`foo` has been deprecated in `v0.1.2`, will be removed ' \
            'in `v1.2.3`. Use of `bar` is recommended instead.'

    def test_deprecation_message(self):
        with warns(DataRobotDeprecationWarning):
            bar(1, 2)


class TestSourcedataUtils(unittest.TestCase):
    def setUp(self):
        self.default_fname = 'predict.csv'

    def test_recognize_sourcedata_passed_dataframe(self):
        df = pd.DataFrame({'a': range(100), 'b': range(100)})
        kwargs = recognize_sourcedata(df, self.default_fname)
        self.assertTrue('filelike' in kwargs)
        self.assertEqual(kwargs.get('fname'), self.default_fname)
        buff = kwargs['filelike']
        readback = pd.read_csv(buff)
        self.assertEqual(readback.columns.tolist(), ['a', 'b'])

    def test_recognize_sourcedata_passed_filelike(self):
        path = fixture_file_path('synthetic-100.csv')
        with open(path, 'rb') as fd:
            kwargs = recognize_sourcedata(fd, self.default_fname)
            self.assertTrue(kwargs.get('filelike') is fd)
            self.assertEqual(kwargs.get('fname'), self.default_fname)

    def test_recognize_sourcedata_passed_filepath(self):
        file_path = fixture_file_path('synthetic-100.csv')
        kwargs = recognize_sourcedata(file_path, self.default_fname)
        self.assertEqual(kwargs.get('file_path'), file_path)
        self.assertEqual(kwargs.get('fname'), 'synthetic-100.csv')

    def test_recognize_sourcedata_passed_content(self):
        content = b'a,b,c\n' + b'1,2,3\n' * 100
        kwargs = recognize_sourcedata(content, self.default_fname)
        self.assertEqual(kwargs.get('content'), content)
        self.assertEqual(kwargs.get('fname'), self.default_fname)

    def test_is_urlsource_passed_true(self):
        result = is_urlsource('http://path_to_urlsource')
        self.assertTrue(result)

    def test_is_urlsource_passed_false(self):
        result = is_urlsource('not_a_path_to_urlsource')
        self.assertFalse(result)

    def test_fatfingered_filepath_raises(self):
        content = b'/home/datarobot/mistypefilepath.csv'
        with self.assertRaises(errors.InputNotUnderstoodError):
            recognize_sourcedata(content, self.default_fname)


def test_resource():
    """Python 3 ResourceWarning of implicitly closed files."""
    with warns():
        open(__file__)


def test_rename_compat():
    """datarobot_sdk rename."""
    with warns(DataRobotDeprecationWarning):
        import datarobot_sdk  # noqa



import trafaret as t

from ..client import get_client, staticproperty
from ..enums import JOB_TYPE, QUEUE_STATUS
from ..utils import from_api


class Job(object):
    """ Generic representation of a job in the queue
    """
    _client = staticproperty(get_client)

    def __init__(self, data):
        self.id = None
        self.project = None
        self.job_type = None
        self.status = None
        self._path = None

        self._data = data
        if isinstance(data, dict):
            self._set_up(data)
        else:
            raise ValueError('job data must be a dict')

    def _set_up(self, data):
        # Importing here to dodge circular dependency - pattern copied from ModelJob/PredictJob
        from . import Project

        data = from_api(data)
        self.project = Project(data.get('project_id'))

        converter = t.Dict({
            t.Key('id', optional=True) >> 'id': t.Int,
            t.Key('status', optional=True) >> 'status': t.String,
            t.Key('project_id', optional=True) >> 'project_id': t.String,
            t.Key('job_type', optional=True) >> 'job_type': t.String,
            t.Key('url') >> 'url': t.String
        }).allow_extra("*")

        try:
            safe_data = converter.check(data)
        except t.DataError as e:
            raise ValueError('Invalid data: {}'.format(str(e)))

        job_type = safe_data.get('job_type')
        if job_type not in [JOB_TYPE.MODEL, JOB_TYPE.PREDICT]:
            raise ValueError('Invalid job_type: {}'.format(job_type))

        status = safe_data.get('status')
        if status not in [QUEUE_STATUS.QUEUE, QUEUE_STATUS.INPROGRESS, QUEUE_STATUS.ERROR]:
            raise ValueError('Invalid status: {}'.format(job_type))

        self.id = safe_data.get('id')
        self.project = Project(safe_data.get('project_id'))
        self.job_type = job_type
        self.status = status
        self._path = self._client.strip_endpoint(safe_data['url'])

    def __repr__(self):
        return 'Job({}, status={})'.format(self.job_type, self.status)

    def _get_full_details(self):
        response = self._client.get(self._path)
        return response.json()

    def cancel(self):
        """ Cancel a running job
        """
        self._client.delete(self._path)



import unittest

from datarobot.helpers import partitioning_methods as pm
from datarobot.helpers.partitioning_methods import BasePartitioningMethod


class TestPartitionMethodDispatch(unittest.TestCase):

    def test_random_cv(self):
        partition_data = dict(
            cv_method='random',
            validation_type='CV',
            holdout_pct=20,
            reps=5
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.RandomCV)

    def test_random_tvh(self):
        partition_data = dict(
            cv_method='random',
            validation_type='TVH',
            holdout_pct=20,
            validation_pct=16
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.RandomTVH)

    def test_stratified_cv(self):
        partition_data = dict(
            cv_method='stratified',
            validation_type='CV',
            holdout_pct=20,
            reps=5
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.StratifiedCV)

    def test_stratified_tvh(self):
        partition_data = dict(
            cv_method='stratified',
            validation_type='TVH',
            holdout_pct=20,
            validation_pct=16
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.StratifiedTVH)

    def test_user_cv(self):
        partition_data = dict(
            cv_method='user',
            validation_type='CV',
            user_partition_col='partitioner',
            cv_holdout_level='holdout_level'
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.UserCV)

    def test_user_tvh(self):
        partition_data = dict(
            cv_method='user',
            validation_type='TVH',
            user_partition_col='partitioner',
            training_level='training',
            validation_level='validation',
            holdout_level='holdout'
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.UserTVH)

    def test_group_cv(self):
        partition_data = dict(
            cv_method='group',
            validation_type='CV',
            partition_key_cols=['groupby1', 'groupby2'],
            holdout_pct=20,
            reps=5,
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.GroupCV)

    def test_group_tvh(self):
        partition_data = dict(
            cv_method='group',
            validation_type='TVH',
            partition_key_cols=['groupby1', 'groupby2'],
            holdout_pct=20,
            validation_pct=16,
        )
        part = BasePartitioningMethod.from_data(partition_data)
        self.assertIsInstance(part, pm.GroupTVH)

    def test_you_screwed_up(self):
        with self.assertRaises(ValueError):
            BasePartitioningMethod.from_data({'bananas': 'and cream'})

    def test_none_in_none_out(self):
        part = BasePartitioningMethod.from_data(None)
        self.assertIsNone(part)



"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

import os
import re
from datetime import datetime, date

import six
from six.moves import configparser
import pandas as pd
from dateutil import parser
import yaml

from .sourcedata import is_urlsource, recognize_sourcedata, dataframe_to_buffer  # noqa
from .deprecation import deprecated, deprecation_warning  # noqa

ALL_CAPITAL = re.compile(r'(.)([A-Z][a-z]+)')
CASE_SWITCH = re.compile(r'([a-z0-9])([A-Z])')
UNDERSCORES = re.compile(r'[a-z]_[a-z]{0,1}')


def underscorize(value):
    partial_result = ALL_CAPITAL.sub(r'\1_\2', value)
    return CASE_SWITCH.sub(r'\1_\2', partial_result).lower()


def underscoreToCamel(match):
    groups = match.group()
    if len(groups) == 2:
        # underscoreToCamel('sample_pct__gte') -> 'samplePct__gte'
        return groups
    return groups[0] + groups[2].upper()


def camelize(value):
    return UNDERSCORES.sub(underscoreToCamel, value)


def from_api(data, do_recursive=True):
    if type(data) not in (dict, list):
        return data
    if isinstance(data, list):
        return _from_api_list(data, do_recursive=do_recursive)
    return _from_api_dict(data, do_recursive=do_recursive)


def _from_api_dict(data, do_recursive=True):
    app_data = {}
    for k, v in six.iteritems(data):
        if v is None:
            continue
        if k == '_id':
            app_data['id'] = v
            continue
        app_data[underscorize(k)] = from_api(v, do_recursive=do_recursive) if do_recursive else v
    return app_data


def _from_api_list(data, do_recursive=True):
    return [from_api(datum, do_recursive=do_recursive) for datum in data]


def remove_empty_keys(metadata):
    return {k: v for k, v in metadata.items() if v is not None}


def parse_time(time_str):
    try:
        return parser.parse(time_str)
    except Exception:
        return time_str


def to_api(data):
    """
    :param data: dictionary {'max_digits': 1}
    :return: {'maxDigits': 1}
    """
    if not data:
        return {}
    assert isinstance(data, dict), 'Wrong type'
    data = remove_empty_keys(data)
    for k, v in six.iteritems(data):
        if isinstance(v, (datetime, date)):
            data[k] = v.isoformat()
    api_data = {camelize(k): to_api(v) if isinstance(v,
                dict) else v for k,
                v in six.iteritems(data)}
    return api_data


def get_id_from_response(response):
    location_string = response.headers['Location']
    return location_string.split('/')[-2]


def get_duplicate_features(features):
    duplicate_features = set()
    seen_features = set()
    for feature in features:
        if feature in seen_features:
            duplicate_features.add(feature)
        else:
            seen_features.add(feature)
    return list(duplicate_features)


def raw_prediction_response_to_dataframe(pred_response):
    """Raw predictions for classification come as nested json objects.

    This will extract it to be a single dataframe.

    Parameters
    ----------
    pred_response : dict
        The loaded json object returned from the prediction route.

    Returns
    -------
    frame : pandas.DataFrame

    """
    predictions = from_api(pred_response)['predictions']
    frame = pd.DataFrame.from_records(predictions)
    return frame


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')
def get_credentials_from_yaml(config_path):
    with open(config_path, "r") as config_file:
        config = yaml.safe_load(config_file)
    username = config.get("username")
    password = config.get("password")
    token = config.get("token")
    return username, password, token


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')
def get_credentials_from_ini(config_path):
    cfg_parser = configparser.RawConfigParser()
    cfg_parser.read(config_path)
    username, password, token = None, None, None
    if cfg_parser.has_option('datarobot', 'username'):
        username = cfg_parser.get('datarobot', 'username')
    if cfg_parser.has_option('datarobot', 'password'):
        password = cfg_parser.get('datarobot', 'password')
    if cfg_parser.has_option('datarobot', 'token'):
        token = cfg_parser.get('datarobot', 'token')
    return username, password, token


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')
def path_is_yaml(path):
    _, file_extension = os.path.splitext(path)
    return file_extension == '.yaml'


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')
def get_credentials_from_file(config_path):
    if path_is_yaml(config_path):
        return get_credentials_from_yaml(config_path)
    else:
        # put deprecation warning here
        # @deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3',
        #     message='Use a YAML configuration file (default location: '
        #             '~/.config/datarobot/drconfig.yaml) instead.')
        return get_credentials_from_ini(config_path)


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')
def get_endpoint_from_yaml(config_path):
    with open(config_path, "r") as config_file:
        config = yaml.safe_load(config_file)
        return config.get('endpoint')


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')
def get_endpoint_from_ini(config_path):
    if config_path:
        cfg_parser = configparser.RawConfigParser()
        cfg_parser.read(config_path)
        try:
            return cfg_parser.get('datarobot', 'endpoint')
        except (configparser.NoSectionError,
                configparser.NoOptionError):
            return None
    return None


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')
def get_endpoint_from_file(config_path):
    if config_path is None:
        return None
    elif path_is_yaml(config_path):
        return get_endpoint_from_yaml(config_path)
    else:
        return get_endpoint_from_ini(config_path)


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')
def get_endpoint():
    ways_of_endpoint = [
        os.environ.get('DATAROBOT_ENDPOINT'),
        get_endpoint_from_file(get_config_file()),
        'https://app.datarobot.com/api/v2']
    return next(endpoint for endpoint in ways_of_endpoint if endpoint)


@deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')
def get_config_file():
    first_choice_config_path = os.path.expanduser('~/.config/datarobot/drconfig.yaml')
    deprecated_config_path = os.path.join(os.path.expanduser('~'), '.datarobotrc')
    if 'DATAROBOT_CONFIG_FILE' in os.environ:
        config_path = os.environ.get('DATAROBOT_CONFIG_FILE')
        if file_exists(config_path):
            return config_path
        else:
            raise ValueError('Bad config path')
    elif file_exists(first_choice_config_path):
        return first_choice_config_path
    elif file_exists(deprecated_config_path):
        return deprecated_config_path
    else:
        return None


file_exists = os.path.isfile
file_exists = deprecated(deprecated_since_version='v2.1', will_remove_version='v2.3')(file_exists)



import collections
import webbrowser

import six
import trafaret as t


from .. import errors
from ..client import get_client, staticproperty
from ..enums import (ASYNC_PROCESS_STATUS, LEADERBOARD_SORT_KEY,
                     AUTOPILOT_MODE, VERBOSITY_LEVEL, QUEUE_STATUS,)
from .job import Job
from .modeljob import ModelJob
from .predict_job import PredictJob
from ..helpers import RecommenderSettings, AdvancedOptions
from ..helpers.partitioning_methods import BasePartitioningMethod
from ..utils import (from_api,
                     parse_time, get_id_from_response,
                     get_duplicate_features, camelize,
                     deprecated,
                     is_urlsource, recognize_sourcedata, retry, logger)
from .feature import Feature

logger = logger.get_logger(__name__)


def _get_target_name(target_obj):
    if isinstance(target_obj, dict):
        return target_obj.get('target_name')
    elif isinstance(target_obj, six.string_types):
        return target_obj


class Project(object):
    _client = staticproperty(get_client)
    _path = 'projects/'
    _fields = frozenset(['id', 'project_name', 'mode', 'target',
                         'target_type', 'holdout_unlocked', 'partition',
                         'metric', 'stage', 'created', 'advanced_options',
                         'recommender', 'positive_class', 'max_train_pct'])

    _converter = t.Dict({
        t.Key('_id', optional=True) >> 'id': t.String(allow_blank=True),
        t.Key('id', optional=True) >> 'id': t.String(allow_blank=True),
        t.Key('project_name', optional=True) >> 'project_name': t.String(
            allow_blank=True),
        t.Key('autopilot_mode', optional=True) >> 'mode': t.Int,
        t.Key('target', optional=True) >> 'target': _get_target_name,
        t.Key('target_type', optional=True): t.String(allow_blank=True),
        t.Key('holdout_unlocked', optional=True): t.Bool(),
        t.Key('metric', optional=True) >> 'metric': t.String(allow_blank=True),
        t.Key('stage', optional=True) >> 'stage': t.String(allow_blank=True),
        t.Key('partition', optional=True): t.Dict().allow_extra('*'),
        t.Key('positive_class', optional=True): t.Or(t.Int(), t.String()),
        t.Key('created', optional=True): parse_time,
        t.Key('advanced_options', optional=True): t.Dict().allow_extra('*'),
        t.Key('recommender', optional=True): t.Dict().allow_extra('*'),
        t.Key('max_train_pct', optional=True): t.Float()
    }).allow_extra('*')

    def __init__(self, data):
        self._data = data
        if isinstance(data, dict):
            self._set_up(data)
        else:
            self.id = data

    def _set_up(self, data):
        data = self._converter.check(from_api(data))
        for k, v in six.iteritems(data):
            if k in self._fields:
                setattr(self, k, v)

    @staticmethod
    def _load_partitioning_method(method, payload):
        if not isinstance(method, BasePartitioningMethod):
            raise TypeError('method should inherit from BasePartitioningMethod')
        payload.update(method.collect_payload())

    @staticmethod
    def _load_recommender_settings(settings, payload):
        if not isinstance(settings, RecommenderSettings):
            raise TypeError('settings should be a RecommenderSettings')
        payload.update(settings.collect_payload())

    @staticmethod
    def _load_advanced_options(opts, payload):
        if not isinstance(opts, AdvancedOptions):
            raise TypeError('opts should inherit from AdvancedOptions')
        payload.update(opts.collect_payload())

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               self.project_name or self.id)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    @classmethod
    def _get_data(cls, project_id):
        """
        Fetch information about a project.

        Used internally when fetching project information to either fetch
        a Project object directly, or update a project with new information
        from the server

        Parameters
        ----------
        project_id : str
            The identifier of the project you want to load.

        Returns
        -------
        project_data : dict
            The queried project's information
        """
        url = '{}{}/'.format(cls._path, project_id)
        return cls._client.get(url).json()

    @classmethod
    def get(cls, project_id):
        """
        Gets information about a project.

        Parameters
        ----------
        project_id : str
            The identifier of the project you want to load.

        Returns
        -------
        project : Project
            The queried project

        Examples
        --------
        .. code-block:: python

            import datarobot as dr
            p = dr.Project.get(project_id='54e639a18bd88f08078ca831')
            p.id
            >>>'54e639a18bd88f08078ca831'
            p.project_name
            >>>'Some project name'
        """
        data = cls._get_data(project_id)
        return cls(data)

    @classmethod
    def create(cls, sourcedata, project_name="Untitled Project", max_wait=600):
        """
        Creates a project with provided data.

        Project creation is asynchronous process, which means that after
        initial request we will keep polling status of async process
        that is responsible for project creation until it's finished.
        For SDK users this only means that this method might raise
        exceptions related to it's async nature.

        Parameters
        ----------
        sourcedata : str, file or pandas.DataFrame
            Data to be used for predictions.
            If str can be either a path to a local file, url to publicly
            available file or raw file content.
        project_name : str, optional
            The name to assign to the empty project.
        max_wait : int, optional
            Time in seconds after which project creation is considered
            unsuccessful

        Returns
        -------
        project : Project
            Instance with initialized data.

        Raises
        ------
        InputNotUnderstoodError
            Raised if `sourcedata` isn't one of supported types.
        AsyncFailureError
            Polling for status of async process resulted in response
            with unsupported status code. Beginning in version 2.1, this
            will be ProjectAsyncFailureError, a subclass of AsyncFailureError
        AsyncProcessUnsuccessfulError
            Raised if project creation was unsuccessful
        AsyncTimeoutError
            Raised if project creation took more time, than specified
            by ``max_wait`` parameter

        Examples
        --------
        .. code-block:: python

            p = Project.create('/home/datasets/somedataset.csv',
                               project_name="New API project")
            p.id
            >>> '5921731dkqshda8yd28h'
            p.project_name
            >>> 'New API project'
        """
        form_data = {"project_name": project_name}

        default_fname = '{}.csv'.format(project_name)

        if is_urlsource(sourcedata):
            form_data["url"] = sourcedata
            initial_project_post_response = cls._client.post(cls._path, data=form_data)
        else:
            filesource_kwargs = recognize_sourcedata(sourcedata, default_fname)
            initial_project_post_response = cls._client.build_request_with_file(
                url=cls._path,
                form_data=form_data,
                method='post',
                **filesource_kwargs)

        async_location = initial_project_post_response.headers['Location']
        return cls.from_async(async_location, max_wait)

    @classmethod
    def from_async(cls, async_location, max_wait=600):
        """
        Given a temporary async status location poll for no more than max_wait seconds
        until the async process (project creation or setting the target, for example)
        finishes successfully, then return the ready project

        Parameters
        ----------
        async_location : str
            The URL for the temporary async status resource. This is returned
            as a header in the response to a request that initiates an
            async process
        max_wait : int
            The maximum number of seconds to wait before giving up.

        Returns
        -------
        project : Project
            The project, now ready

        Raises
        ------
        ProjectAsyncFailureError
            If the server returned an unexpected response while polling for the
            asynchronous operation to resolve
        AsyncProcessUnsuccessfulError
            If the final result of the asynchronous operation was a failure
        AsyncTimeoutError
            If the asynchronous operation did not resolve within the time
            specified
        """
        try:
            for _ in retry.wait(max_wait):
                r = cls._client.get(async_location, join_endpoint=False, allow_redirects=False)
                if r.status_code == 303:
                    project_id = get_id_from_response(r)
                    return cls.get(project_id)
                response_data = r.json()
                status = response_data.get('status')
                if status == ASYNC_PROCESS_STATUS.ERROR or \
                        status == ASYNC_PROCESS_STATUS.ABORTED:
                    e_msg = 'Async process unsuccessful: {}'
                    raise errors.AsyncProcessUnsuccessfulError(
                        e_msg.format(response_data.get('message')))

            timeout_msg = (
                'Status did not reach 303 - it timed out in {} seconds; '
                'last status was {}: {}'
            ).format(max_wait, r.status_code, r.text)
            raise errors.AsyncTimeoutError(timeout_msg)

        except errors.AppPlatformError as e:
            raise errors.ProjectAsyncFailureError(repr(e), e.status_code, async_location)

    @classmethod
    def start(cls, sourcedata,
              target,
              project_name='Untitled Project',
              worker_count=None,
              metric=None,
              autopilot_on=True,
              recommender_settings=None,
              blueprint_threshold=None,
              response_cap=None,
              partitioning_method=None,
              positive_class=None,
              ):
        """
        Chain together project creation, file upload, and target selection.

        Parameters
        ----------
        sourcedata : str or pandas.DataFrame
            The path to the file to upload. Can be either a path to a
            local file or a publicly accessible URL.
            If the source is a DataFrame, it will be serialized to a
            temporary buffer.
        target : str
            The name of the target column in the uploaded file.
        project_name : str
            The project name.

        Other Parameters
        ----------------
        worker_count : int, optional
            The number of workers that you want to allocate to this project.
        metric : str, optional
            The name of metric to use.
        autopilot_on : boolean, default ``True``
            Whether or not to begin modeling automatically.
        recommender_settings : RecommenderSettings, optional
            Columns specified in this object tell the
            system how to set up the recommender system
        blueprint_threshold : int, optional
            Number of hours the model is permitted to run.
            Minimum 1
        response_cap : float, optional
            Quantile of the response distribution to use for response capping
            Must be in range 0.5 .. 1.0
        partitioning_method : PartitioningMethod object, optional
            It should be one of PartitioningMethod object.
        positive_class : str, float, or int; optional
            Specifies a level of the target column that should be used for
            binary classification. Can be used to specify any of the available
            levels as the binary target - all other levels will be treated
            as a single negative class.

        Returns
        -------
        project : Project
            The newly created and initialized project.

        Raises
        ------
        AsyncFailureError
            Polling for status of async process resulted in response
            with unsupported status code
        AsyncProcessUnsuccessfulError
            Raised if project creation or target setting was unsuccessful
        AsyncTimeoutError
            Raised if project creation or target setting timed out

        Examples
        --------

        .. code-block:: python

            Project.start("./tests/fixtures/file.csv",
                          "a_target",
                          project_name="test_name",
                          worker_count=4,
                          metric="a_metric")
        """

        # Create project part
        create_data = {'project_name': project_name, 'sourcedata': sourcedata}
        project = cls.create(**create_data)

        # Set target
        if autopilot_on:
            mode = AUTOPILOT_MODE.FULL_AUTO
        else:
            mode = AUTOPILOT_MODE.MANUAL

        advanced_options = AdvancedOptions(
            blueprint_threshold=blueprint_threshold,
            response_cap=response_cap
        )

        project.set_target(
            target, metric=metric, mode=mode,
            worker_count=worker_count,
            recommender_settings=recommender_settings,
            advanced_options=advanced_options,
            partitioning_method=partitioning_method,
            positive_class=positive_class)
        return project

    @classmethod
    def list(cls, search_params=None):
        """
        Returns the projects associated with this account.

        Parameters
        ----------
        search_params : dict, optional.
            If not `None`, the returned projects are filtered by lookup.
            Currently you can query projects by:

            * ``project_name``

        Returns
        -------
        projects : list of Project instances
            Contains a list of projects associated with this user
            account.

        Raises
        ------
        ValueError
            Raised if ``search_params`` parameter is provided,
            but is not of supported type.

        Examples
        --------
        List all projects
        .. code-block:: python

            p_list = Project.list()
            p_list
            >>> [Project('Project One'), Project('Two')]

        Search for projects by name
        .. code-block:: python

            Project.list(search_params={'project_name': 'red'})
            >>> [Project('Predtime'), Project('Fred Project')]

        """
        get_params = {}
        if search_params is not None:
            if isinstance(search_params, dict):
                get_params.update(search_params)
            else:
                raise TypeError(
                    'Provided search_params argument {} is invalid type {}'.format(
                        search_params, type(search_params)
                    ))
        r_data = cls._client.get(cls._path, params=get_params).json()
        return [cls(item) for item in r_data]

    @deprecated(deprecated_since_version='v2.2',
                will_remove_version='v2.4',
                message='Please use `rename`, `unlock_holdout`, or `set_worker_count` instead.')
    def update(self, **data):
        """
        Change the project properties.

        Other Parameters
        ----------------
        project_name : str, optional
            The name to assign to this project.

        holdout_unlocked : bool, optional
            Can only have value of `True`. If
            passed, unlocks holdout for project.

        worker_count : int, optional
            Sets number or workers

        Returns
        -------
        project : Project
            Instance with fields updated.
        """
        return self._update(**data)

    def _update(self, **data):
        """
        Change the project properties.

        In the future, DataRobot API will provide endpoints to directly
        update the attributes currently handled by this one endpoint.

        Other Parameters
        ----------------
        project_name : str, optional
            The name to assign to this project.

        holdout_unlocked : bool, optional
            Can only have value of `True`. If
            passed, unlocks holdout for project.

        worker_count : int, optional
            Sets number or workers

        Returns
        -------
        project : Project
            Instance with fields updated.
        """
        for key in (set(data) - {'project_name', 'holdout_unlocked', 'worker_count'}):
            raise TypeError("update() got an unexpected keyword argument '{}'".format(key))
        url = '{}{}/'.format(self._path, self.id)
        self._client.patch(url, data=data)
        self._set_up(data)
        return self

    def delete(self):
        """Removes this project from your account.
        """
        url = '{}{}/'.format(self._path, self.id)
        self._client.delete(url)

    def set_target(self, target,
                   mode=AUTOPILOT_MODE.FULL_AUTO,
                   metric=None,
                   quickrun=None,
                   worker_count=None,
                   positive_class=None,
                   recommender_settings=None,
                   partitioning_method=None,
                   featurelist_id=None,
                   advanced_options=None,
                   max_wait=600,
                   ):
        """
        Set target variable of an existing project that has a file uploaded
        to it.

        Target setting is asynchronous process, which means that after
        initial request we will keep polling status of async process
        that is responsible for target setting until it's finished.
        For SDK users this only means that this method might raise
        exceptions related to it's async nature.

        Parameters
        ----------
        target : str
            Name of variable.
        mode : int, optional
            You can use ``AUTOPILOT_MODE`` enum to choose between

            * ``AUTOPILOT_MODE.FULL_AUTO``
            * ``AUTOPILOT_MODE.SEMI_AUTO``
            * ``AUTOPILOT_MODE.MANUAL``

            If unspecified, ``FULL_AUTO`` is used
        metric : str, optional
            Name of the metric to use for evaluating models. You can query
            the metrics available for the target by way of
            ``Project.get_metrics``. If none is specified, then the default
            recommended by DataRobot is used.
        quickrun : bool, optional
            Sets whether project should be run in ``quick run`` mode. This
            setting causes DataRobot to recommend a more limited set of models
            in order to get a base set of models and insights more quickly.
        worker_count : int, optional
            The number of concurrent workers to request for this project. If
            `None`, then the default is used
        recommender_settings : RecommenderSettings, optional
            if not `None`, then the columns specified in this object tell the
            system how to set up the recommender system
        partitioning_method : PartitioningMethod object, optional
            It should be one of PartitioningMethod object.
        positive_class : str, float, or int; optional
            Specifies a level of the target column that treated as the
            positive class for binary classification.  May only be specified
            for binary classification targets.
        featurelist_id : str, optional
            Specifies which feature list to use.
        advanced_options : AdvancedOptions, optional
            Used to set advanced options of project creation.
        max_wait : int, optional
            Time in seconds after which target setting is considered
            unsuccessful.

        Returns
        -------
        project : Project
            The instance with updated attributes.

        Raises
        ------
        AsyncFailureError
            Polling for status of async process resulted in response
            with unsupported status code
        AsyncProcessUnsuccessfulError
            Raised if target setting was unsuccessful
        AsyncTimeoutError
            Raised if target setting took more time, than specified
            by ``max_wait`` parameter

        See Also
        --------
        Project.start : combines project creation, file upload, and target
            selection
        """
        if worker_count is not None:
            self.set_worker_count(worker_count)

        aim_payload = {
            'target': target,
            'mode': mode,
            'metric': metric,
        }
        if recommender_settings:
            self._load_recommender_settings(
                recommender_settings, aim_payload)
        if advanced_options is not None:
            self._load_advanced_options(
                advanced_options, aim_payload)
        if positive_class is not None:
            aim_payload['positive_class'] = positive_class
        if quickrun is not None:
            aim_payload['quickrun'] = quickrun
        if featurelist_id is not None:
            aim_payload['featurelist_id'] = featurelist_id
        if partitioning_method:
            self._load_partitioning_method(partitioning_method, aim_payload)
        url = '{}{}/aim/'.format(self._path, self.id)
        response = self._client.patch(url, data=aim_payload)
        async_location = response.headers['Location']

        # Waits for project to be ready for modeling, but ignores the return value
        self.from_async(async_location, max_wait=max_wait)

        proj_data = self._get_data(self.id)
        self._set_up(proj_data)
        return self

    def get_models(self, order_by=None, search_params=None, with_metric=None):
        """
        List all completed, successful models in the leaderboard for the given project.

        Parameters
        ----------
        order_by : str or list of strings, optional
            If not `None`, the returned models are ordered by this
            attribute. If `None`, the default return is the order of
            default project metric.

            Allowed attributes to sort by are:

            * ``metric``
            * ``sample_pct``

            If the sort attribute is preceded by a hyphen, models will be sorted in descending
            order, otherwise in ascending order.

            Multiple sort attributes can be included as a comma-delimited string or in a list
            e.g. order_by=`sample_pct,-metric` or order_by=[`sample_pct`, `-metric`]

            Using `metric` to sort by will result in models being sorted according to their
            validation score by how well they did according to the project metric.
        search_params : dict, optional.
            If not `None`, the returned models are filtered by lookup.
            Currently you can query models by:

            * ``name``
            * ``sample_pct``

        with_metric : str, optional.
            If not `None`, the returned models will only have scores for this
            metric. Otherwise all the metrics are returned.

        Returns
        -------
        models : a list of Model instances.
            All of the models that have been trained in this project.

        Raises
        ------
        ValueError
            Raised if ``order_by`` or ``search_params`` parameter is provided,
            but is not of supported type.

        Examples
        --------

        .. code-block:: python

            Project('pid').get_models(order_by=['-sample_pct',
                                                'metric'])

            # Getting models that contain "Ridge" in name
            # and with sample_pct more than 64
            Project('pid').get_models(
                search_params={
                    'sample_pct__gt': 64,
                    'name': "Ridge"
                })
        """
        from . import Model

        url = '{}{}/models/'.format(self._path, self.id)
        get_params = {}
        if order_by is not None:
            order_by = self._canonize_order_by(order_by)
            get_params.update({'order_by': order_by})
        else:
            get_params.update({'order_by': '-metric'})
        if search_params is not None:
            if isinstance(search_params, dict):
                get_params.update(search_params)
            else:
                raise TypeError(
                    'Provided search_params argument is invalid')
        if with_metric is not None:
            get_params.update({'with_metric': with_metric})
        resp_data = self._client.get(url, params=get_params).json()
        return [Model(item, project=self) for item in resp_data]

    def _canonize_order_by(self, order_by):
        legal_keys = [
            LEADERBOARD_SORT_KEY.SAMPLE_PCT,
            LEADERBOARD_SORT_KEY.PROJECT_METRIC,
        ]
        processed_keys = []
        if order_by is None:
            return order_by
        if isinstance(order_by, str):
            order_by = order_by.split(',')
        if not isinstance(order_by, list):
            msg = 'Provided order_by attribute {} is of an unsupported type'.format(order_by)
            raise TypeError(msg)
        for key in order_by:
            key = key.strip()
            if key.startswith('-'):
                prefix = '-'
                key = key[1:]
            else:
                prefix = ''
            if key not in legal_keys:
                camel_key = camelize(key)
                if camel_key not in legal_keys:
                    msg = 'Provided order_by attribute {}{} is invalid'.format(prefix, key)
                    raise ValueError(msg)
                key = camel_key
            processed_keys.append('{}{}'.format(prefix, key))
        return ','.join(processed_keys)

    def get_blueprints(self):
        """
        List all blueprints recommended for a project.

        Returns
        -------
        menu : list of Blueprint instances
            All the blueprints recommended by DataRobot for a project
        """
        from . import Blueprint

        url = '{}{}/blueprints/'.format(self._path, self.id)
        resp_data = self._client.get(url).json()
        return [Blueprint(item) for item in resp_data]

    def get_features(self):
        """
        List all features for this project

        Returns
        -------
        list of Features
            all feature for this project
        """
        url = '{}{}/features/'.format(self._path, self.id)
        resp_data = self._client.get(url).json()
        return [Feature(item) for item in resp_data]

    def get_featurelists(self):
        """
        List all featurelists created for this project

        Returns
        -------
        list of Featurelist
            all featurelists created for this project
        """
        from . import Featurelist

        url = '{}{}/featurelists/'.format(self._path, self.id)
        resp_data = self._client.get(url).json()
        return [Featurelist(item) for item in resp_data]

    def create_featurelist(self, name, features):
        """
        Creates a new featurelist

        Parameters
        ----------
        name : str
            The name to give to this new featurelist. Names must be unique, so
            an error will be returned from the server if this name has already
            been used in this project.
        features : list of str
            The names of the features. Each feature must exist in the project
            already.

        Returns
        -------
        Featurelist
            newly created featurelist

        Raises
        ------
        DuplicateFeaturesError
            Raised if `features` variable contains duplicate features

        Examples
        --------
        .. code-block:: python

            project = Project('5223deadbeefdeadbeef0101')
            flists = project.get_featurelists()

            # Create a new featurelist using a subset of features from an
            # existing featurelist
            flist = flists[0]
            features = flist.features[::2]  # Half of the features

            new_flist = project.create_featurelist(name='Feature Subset',
                                                   features=features)
        """
        from . import Featurelist

        url = '{}{}/featurelists/'.format(self._path, self.id)

        duplicate_features = get_duplicate_features(features)
        if duplicate_features:
            err_msg = "Can't create featurelist with duplicate " \
                      "features - {}".format(duplicate_features)
            raise errors.DuplicateFeaturesError(err_msg)

        payload = {
            'name': name,
            'features': features,
        }
        response = self._client.post(url, data=payload)
        new_id = get_id_from_response(response)
        flist_data = {'name': name,
                      'id': new_id,
                      'features': features,
                      'project_id': self.id}
        return Featurelist(flist_data)

    def get_metrics(self, feature_name):
        """Get the metrics recommended for modeling on the given feature.

        Parameters
        ----------
        feature_name : str
            The name of the feature to query regarding which metrics are
            recommended for modeling.

        Returns
        -------
        names : list of str
            The names of the recommended metrics.
        """
        url = '{}{}/features/metrics/'.format(self._path, self.id)
        params = {
            'feature_name': feature_name
        }
        return from_api(self._client.get(url, params=params).json())

    def get_status(self):
        """Query the server for project status.

        Returns
        -------
        status : dict
            Contains:

            * ``autopilot_done`` : a boolean.
            * ``stage`` : a short string indicating which stage the project
              is in.
            * ``stage_description`` : a description of what ``stage`` means.

        Examples
        --------

        .. code-block:: python

            {"autopilot_done": False,
             "stage": "modeling",
             "stage_description": "Ready for modeling"}
        """
        url = '{}{}/status/'.format(self._path, self.id)
        return from_api(self._client.get(url).json())

    def pause_autopilot(self):
        """
        Pause autopilot, which stops processing the next jobs in the queue.

        Returns
        -------
        paused : boolean
            Whether the command was acknowledged
        """
        url = '{}{}/autopilot/'.format(self._path, self.id)
        payload = {
            'command': 'stop'
        }
        self._client.post(url, data=payload)

        return True

    def unpause_autopilot(self):
        """
        Unpause autopilot, which restarts processing the next jobs in the queue.

        Returns
        -------
        unpaused : boolean
            Whether the command was acknowledged.
        """
        url = '{}{}/autopilot/'.format(self._path, self.id)
        payload = {
            'command': 'start',
        }
        self._client.post(url, data=payload)
        return True

    def start_autopilot(self, featurelist_id, mode=AUTOPILOT_MODE.FULL_AUTO):
        """Starts autopilot on provided featurelist.

        Only one autopilot can be running at the time.
        That's why any ongoing autopilot on a different featurelist will
        be halted - modeling jobs in queue would not
        be affected but new jobs would not be added to queue by
        the halted autopilot.

        Parameters
        ----------
        featurelist_id : str
            Identifier of featurelist that should be used for autopilot
        mode : str, optional
            Mode in which to run autopilot. You can use ``AUTOPILOT_MODE``
            enum to choose between

            * ``AUTOPILOT_MODE.FULL_AUTO``
            * ``AUTOPILOT_MODE.SEMI_AUTO``

            If unspecified, ``AUTOPILOT_MODE.FULL_AUTO`` is used

        Raises
        ------
        AppPlatformError
            Raised if autopilot is currently running on or has already
            finished running on the provided featurelist.
            Also raised if project's target was not selected.
        """
        url = '{}{}/autopilots/'.format(self._path, self.id)
        payload = {
            'featurelistId': featurelist_id,
            'mode': mode,
        }
        self._client.post(url, data=payload)

    def train(self, trainable, sample_pct=None, featurelist_id=None,
              source_project_id=None, scoring_type=None):
        """Submit a job to the queue.

        Parameters
        ----------
        trainable : str or Blueprint
            For ``str``, this is assumed to be a blueprint_id. If no
            ``source_project_id`` is provided, the ``project_id`` will be assumed
            to be the project that this instance represents.

            Otherwise, for a ``Blueprint``, it contains the
            blueprint_id and source_project_id that we want
            to use. ``featurelist_id`` will assume the default for this project
            if not provided, and ``sample_pct`` will default to using the maximum
            training value allowed for this project's partition setup.
            ``source_project_id`` will be ignored if a
            ``Blueprint`` instance is used for this parameter
        sample_pct : float, optional
            The amount of training data to use. Defaults to the maximum
            amount available based on the project configuration.
        featurelist_id : str, optional
            The identifier of the featurelist to use. If not defined, the
            default for this project is used.
        source_project_id : str, optional
            Which project created this blueprint_id. If ``None``, it defaults
            to looking in this project. Note that you must have read
            permissions in this project.
        scoring_type : str, optional
            Either ``SCORING_TYPE.validation`` or
            ``SCORING_TYPE.cross_validation``. ``SCORING_TYPE.validation``
            is available for every partitioning type, and indicates that
            the default model validation should be used for the project.
            If the project uses a form of cross-validation partitioning,
            ``SCORING_TYPE.cross_validation`` can also be used to indicate
            that all of the available training/validation combinations
            should be used to evaluate the model.

        Returns
        -------
        model_job_id : str
            id of created job, can be used as parameter to ``ModelJob.get``
            method or ``wait_for_async_model_creation`` function

        Examples
        --------
        Use a ``Blueprint`` instance:

        .. code-block:: python

            blueprint = project.get_blueprints()[0]
            model_job_id = project.train(blueprint, sample_pct=64)

        Use a ``blueprint_id``, which is a string. In the first case, it is
        assumed that the blueprint was created by this project. If you are
        using a blueprint used by another project, you will need to pass the
        id of that other project as well.

        .. code-block:: python

            blueprint_id = 'e1c7fc29ba2e612a72272324b8a842af'
            project.train(blueprint, sample_pct=64)

            another_project.train(blueprint, source_project_id=project.id)

       You can also easily use this interface to train a new model using the data from
       an existing model:

       .. code-block:: python

           model = project.get_models()[0]
           model_job_id = project.train(model.blueprint.id,
                                        sample_pct=100)

        """
        try:
            return self._train(trainable.id,
                               featurelist_id=featurelist_id,
                               source_project_id=trainable.project_id,
                               sample_pct=sample_pct,
                               scoring_type=scoring_type)
        except AttributeError:
            return self._train(trainable,
                               featurelist_id=featurelist_id,
                               source_project_id=source_project_id,
                               sample_pct=sample_pct,
                               scoring_type=scoring_type)

    def _train(self,
               blueprint_id,
               featurelist_id=None,
               source_project_id=None,
               sample_pct=None,
               scoring_type=None):
        """
        Submit a modeling job to the queue. Upon success, the new job will
        be added to the end of the queue.

        Parameters
        ----------
        blueprint_id: str
            The id of the model. See ``Project.get_blueprints`` to get the list
            of all available blueprints for a project.
        featurelist_id: str, optional
            The dataset to use in training. If not specified, the default
            dataset for this project is used.
        source_project_id : str, optional
            Which project created this blueprint_id. If ``None``, it defaults
            to looking in this project. Note that you must have read
            permisisons in this project.
        sample_pct: float, optional
            The amount of training data to use. Defaults to the maximum
            amount available based on the project configuration.

        Returns
        -------
        model_job_id : str
            id of created job, can be used as parameter to ``ModelJob.get``
            method or ``wait_for_async_model_creation`` function
        """
        url = '{}{}/models/'.format(self._path, self.id)
        payload = {'blueprint_id': blueprint_id}
        if featurelist_id is not None:
            payload['featurelist_id'] = featurelist_id
        if source_project_id is not None:
            payload['source_project_id'] = source_project_id
        if sample_pct is not None:
            payload['sample_pct'] = sample_pct
        if scoring_type is not None:
            payload['scoring_type'] = scoring_type
        response = self._client.post(url, data=payload)
        return get_id_from_response(response)

    def get_all_jobs(self, status=None):
        """Get a list of jobs

        Parameters
        ----------
        status : QUEUE_STATUS enum, optional
            If called with QUEUE_STATUS.INPROGRESS, will return the modeling jobs
            that are currently running.

            If called with QUEUE_STATUS.QUEUE, will return the modeling jobs that
            are waiting to be run.

            If called with QUEUE_STATUS.ERROR, will return the modeling jobs that
            have errored.

            If no value is provided, will return all modeling jobs currently running
            or waiting to be run.

        Returns
        -------
        jobs : list
            Each is an instance of Job
        """
        url = '{}{}/jobs/'.format(self._path, self.id)
        params = {'status': status}
        res = self._client.get(url, params=params).json()
        return [Job(item) for item in res['jobs']]

    @deprecated('v2.1', 'v2.3', message='Please use `get_model_jobs` instead.')
    def get_jobs(self, status=None):
        return self.get_model_jobs(status=status)

    def get_model_jobs(self, status=None):
        """Get a list of modeling jobs

        Parameters
        ----------
        status : QUEUE_STATUS enum, optional
            If called with QUEUE_STATUS.INPROGRESS, will return the modeling jobs
            that are currently running.

            If called with QUEUE_STATUS.QUEUE, will return the modeling jobs that
            are waiting to be run.

            If called with QUEUE_STATUS.ERROR, will return the modeling jobs that
            have errored.

            If no value is provided, will return all modeling jobs currently running
            or waiting to be run.

        Returns
        -------
        jobs : list
            Each is an instance of ModelJob
        """
        url = '{}{}/modelJobs/'.format(self._path, self.id)
        params = {'status': status}
        res = self._client.get(url, params=params).json()
        return [ModelJob(item) for item in res]

    def get_predict_jobs(self, status=None):
        """Get a list of prediction jobs

        Parameters
        ----------
        status : QUEUE_STATUS enum, optional
            If called with QUEUE_STATUS.INPROGRESS, will return the prediction jobs
            that are currently running.

            If called with QUEUE_STATUS.QUEUE, will return the prediction jobs that
            are waiting to be run.

            If called with QUEUE_STATUS.ERROR, will return the prediction jobs that
            have errored.

            If called without a status, will return all prediction jobs currently running
            or waiting to be run.

        Returns
        -------
        jobs : list
            Each is an instance of PredictJob
        """
        url = '{}{}/predictJobs/'.format(self._path, self.id)
        params = {'status': status}
        res = self._client.get(url, params=params).json()
        return [PredictJob(item) for item in res]

    def _get_job_status_counts(self):
        jobs = self.get_model_jobs()
        job_counts = collections.Counter(job.status for job in jobs)
        return job_counts[QUEUE_STATUS.INPROGRESS], job_counts[QUEUE_STATUS.QUEUE]

    def wait_for_autopilot(self, check_interval=20.0, timeout=24*60*60, verbosity=1):
        """
        Blocks until autopilot is finished.

        Parameters
        ----------
        check_interval : float or int
            The maximum time (in seconds) to wait between checks for whether autopilot is finished
        timeout : float or int or None
            After this long (in seconds), we give up. If None, never timeout.
        verbosity:
            This should be VERBOSITY_LEVEL.SILENT or VERBOSITY_LEVEL.VERBOSE.
            For VERBOSITY_LEVEL.SILENT, nothing will be displayed about progress.
            For VERBOSITY_LEVEL.VERBOSE, the number of jobs in progress or queued is shown.
            Note that new jobs are added to the queue along the way.
        """
        for _, seconds_waited in retry.wait(timeout, maxdelay=check_interval):
            if verbosity > VERBOSITY_LEVEL.SILENT:
                num_inprogress, num_queued = self._get_job_status_counts()
                logger.info("In progress: {0}, queued: {1} (waited: {2:.0f}s)".
                            format(num_inprogress, num_queued, seconds_waited))
            status = self.get_status()
            if status['autopilot_done']:
                return
        raise errors.AsyncTimeoutError('Autopilot did not finish within timeout period')

    def rename(self, project_name):
        """Update the name of the project.

        Parameters
        ----------
        project_name : str
            The new name
        """
        self._update(project_name=project_name)

    def unlock_holdout(self):
        """Unlock the holdout for this project.

        This will cause subsequent queries of the models of this project to
        contain the metric values for the holdout set, if it exists.

        Take care, as this cannot be undone. Remember that best practice is to
        select a model before analyzing the model performance on the holdout set
        """
        return self._update(holdout_unlocked=True)

    def set_worker_count(self, worker_count):
        """Sets the number of workers allocated to this project.

        Note that this value is limited to the number allowed by your account.
        Lowering the number will not stop currently running jobs, but will
        cause the queue to wait for the appropriate number of jobs to finish
        before attempting to run more jobs.

        Parameters
        ----------
        worker_count : int
            The number of concurrent workers to request from the pool of workers
        """
        return self._update(worker_count=worker_count)

    def get_leaderboard_ui_permalink(self):
        """
        Returns
        -------
        url : str
            Permanent static hyperlink to a project leaderboard.
        """
        return '{}/{}{}/models'.format(self._client.domain, self._path, self.id)

    def open_leaderboard_browser(self):
        """
        Opens project leaderboard in web browser.

        Note:
        If text-mode browsers are used, the calling process will block
        until the user exits the browser.
        """

        url = self.get_leaderboard_ui_permalink()
        return webbrowser.open(url)



import json
import responses

from datarobot import Project, AppPlatformError
from .utils import SDKTestcase


class Test400LevelErrors(SDKTestcase):

    @responses.activate
    def test_401_error_message(self):
        """For whatever reason, the `requests` library completely ignores the
        body in a 401 error, so, here is our own message
        """
        responses.add(
            responses.GET,
            'https://host_name.com/projects/',
            status=401,
        )

        with self.assertRaisesRegexp(AppPlatformError,
                                     'not properly authenticated'):
            Project.list()

    @responses.activate
    def test_403_error_message(self):
        """Same deal here, the `requests` library completely ignores the
        body in a 403 error, so, here is our own message
        """
        responses.add(
            responses.GET,
            'https://host_name.com/projects/',
            status=403,
        )

        with self.assertRaisesRegexp(AppPlatformError,
                                     'permissions'):
            Project.list()

    @responses.activate
    def test_extracts_error_message(self):
        data = {'message': 'project p-id has been deleted'}
        responses.add(responses.GET, 'https://host_name.com/projects/p-id/models/',
                      status=422,
                      body=json.dumps(data),
                      content_type='application/json')

        with self.assertRaisesRegexp(AppPlatformError, 'project p-id has been deleted'):
            Project({'id': 'p-id'}).get_models()



__all__ = (
    'RandomCV',
    'StratifiedCV',
    'GroupCV',
    'UserCV',
    'RandomTVH',
    'UserTVH',
    'StratifiedTVH',
    'GroupTVH'
)


def get_class(cv_method, validation_type):
    classes = {
        'CV': {
            'random': RandomCV,
            'stratified': StratifiedCV,
            'user': UserCV,
            'group': GroupCV,
        },
        'TVH': {
            'random': RandomTVH,
            'stratified': StratifiedTVH,
            'user': UserTVH,
            'group': GroupTVH,
        },
    }
    try:
        return classes[validation_type][cv_method]
    except KeyError:
        err_msg = 'Error in getting class for cv_method={} and validation_type={}'
        raise ValueError(err_msg.format(cv_method, validation_type))


class BasePartitioningMethod(object):

    """This is base class to describe partioning method
    with options"""

    cv_method = None
    validation_type = None
    seed = 0
    _data = None
    _static_fields = frozenset(['cv_method', 'validation_type'])

    def __init__(self, cv_method, validation_type, seed=0):
        self.cv_method = cv_method
        self.validation_type = validation_type
        self.seed = seed

    def collect_payload(self):
        """
        This method is should return dict that
        will be passed into request to datarobot cloud
        """
        keys = ('cv_method', 'validation_type', 'reps', 'user_partition_col',
                'training_level', 'validation_level', 'holdout_level', 'cv_holdout_level',
                'seed', 'validation_pct', 'holdout_pct', 'partition_key_cols')
        if not self._data:
            self._data = {key: getattr(self, key, None) for key in keys}
        return self._data

    def __repr__(self):
        if self._data:
            payload = {k: v for k, v in self._data.items()
                       if v is not None and k not in self._static_fields}
        else:
            self.collect_payload()
            return repr(self)
        return '{}({})'.format(self.__class__.__name__, payload)

    @classmethod
    def from_data(cls, data):
        """Can be used to instantiate the correct class of partitioning class
        based on the data
        """
        if data is None:
            return None
        cv_method = data.get('cv_method')
        validation_type = data.get('validation_type')
        other_params = {key: value for key, value in data.items()
                        if key not in ['cv_method', 'validation_type']}
        return get_class(cv_method, validation_type)(**other_params)


class BaseCrossValidation(BasePartitioningMethod):
    cv_method = None
    validation_type = 'CV'

    def __init__(self, cv_method, validation_type='CV'):
        self.cv_method = cv_method  # pragma: no cover
        self.validation_type = validation_type  # pragma: no cover


class BaseTVH(BasePartitioningMethod):
    cv_method = None
    validation_type = 'TVH'

    def __init__(self, cv_method, validation_type='TVH'):
        self.cv_method = cv_method  # pragma: no cover
        self.validation_type = validation_type  # pragma: no cover


class RandomCV(BaseCrossValidation):
    """A partition in which observations are randomly assigned to cross-validation groups
    and the holdout set.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    reps : int
        number of cross validation folds to use
    seed : int
        a seed to use for randomization
    """
    cv_method = 'random'

    def __init__(self, holdout_pct, reps, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.reps = reps  # pragma: no cover
        self.seed = seed  # pragma: no cover


class StratifiedCV(BaseCrossValidation):
    """A partition in which observations are randomly assigned to cross-validation groups
    and the holdout set, preserving in each group the same ratio of positive to negative cases as in
    the original data.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    reps : int
        number of cross validation folds to use
    seed : int
        a seed to use for randomization
    """
    cv_method = 'stratified'

    def __init__(self, holdout_pct, reps, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.reps = reps  # pragma: no cover
        self.seed = seed  # pragma: no cover


class GroupCV(BaseCrossValidation):
    """ A partition in which one column is specified, and rows sharing a common value
    for that column are guaranteed to stay together in the partitioning into cross-validation
    groups and the holdout set.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    reps : int
        number of cross validation folds to use
    partition_key_cols : list
        a list containing a single string, where the string is the name of the column whose
        values should remain together in partitioning
    seed : int
        a seed to use for randomization
    """
    cv_method = 'group'

    def __init__(self, holdout_pct, reps, partition_key_cols, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.reps = reps  # pragma: no cover
        self.partition_key_cols = partition_key_cols  # pragma: no cover
        self.seed = seed  # pragma: no cover


class UserCV(BaseCrossValidation):
    """ A partition where the cross-validation folds and the holdout set are specified by
    the user.

    Parameters
    ----------
    user_partition_col : string
        the name of the column containing the partition assignments
    cv_holdout_level
        the value of the partition column indicating a row is part of the holdout set
    seed : int
        a seed to use for randomization
    """
    cv_method = 'user'

    def __init__(self, user_partition_col, cv_holdout_level, seed=0):
        self.user_partition_col = user_partition_col  # pragma: no cover
        self.cv_holdout_level = cv_holdout_level  # pragma: no cover
        self.seed = seed  # pragma: no cover


class RandomTVH(BaseTVH):
    """Specifies a partitioning method in which rows are randomly assigned to training, validation,
    and holdout.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    validation_pct : int
        the desired percentage of dataset to assign to validation set
    seed : int
        a seed to use for randomization
    """
    cv_method = 'random'

    def __init__(self, holdout_pct, validation_pct, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.validation_pct = validation_pct  # pragma: no cover
        self.seed = seed  # pragma: no cover


class UserTVH(BaseTVH):
    """Specifies a partitioning method in which rows are assigned by the user to training,
    validation, and holdout sets.

    Parameters
    ----------
    user_partition_col : string
        the name of the column containing the partition assignments
    training_level
        the value of the partition column indicating a row is part of the training set
    validation_level
        the value of the partition column indicating a row is part of the validation set
    holdout_level
        the value of the partition column indicating a row is part of the holdout set (use
        None if you want no holdout set)
    seed : int
        a seed to use for randomization
    """
    cv_method = 'user'

    def __init__(self, user_partition_col, training_level, validation_level,
                 holdout_level, seed=0):
        self.user_partition_col = user_partition_col  # pragma: no cover
        self.training_level = training_level  # pragma: no cover
        self.validation_level = validation_level  # pragma: no cover
        self.holdout_level = holdout_level  # pragma: no cover
        self.seed = seed  # pragma: no cover


class StratifiedTVH(BaseTVH):
    """A partition in which observations are randomly assigned to train, validation, and
    holdout sets, preserving in each group the same ratio of positive to negative cases as in the
    original data.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    validation_pct : int
        the desired percentage of dataset to assign to validation set
    seed : int
        a seed to use for randomization
    """
    cv_method = 'stratified'

    def __init__(self, holdout_pct, validation_pct, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.validation_pct = validation_pct  # pragma: no cover
        self.seed = seed  # pragma: no cover


class GroupTVH(BaseTVH):
    """A partition in which one column is specified, and rows sharing a common value
    for that column are guaranteed to stay together in the partitioning into the training,
    validation, and holdout sets.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    validation_pct : int
        the desired percentage of dataset to assign to validation set
    partition_key_cols : list
        a list containing a single string, where the string is the name of the column whose
        values should remain together in partitioning
    seed : int
        a seed to use for randomization
    """
    cv_method = 'group'

    def __init__(self, holdout_pct, validation_pct, partition_key_cols,
                 seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.validation_pct = validation_pct  # pragma: no cover
        self.partition_key_cols = partition_key_cols  # pragma: no cover
        self.seed = seed  # pragma: no cover



import six
import trafaret as t

from .. import errors
from ..client import get_client, staticproperty
from ..enums import JOB_TYPE, PREDICT_JOB_STATUS
from .job import Job
from ..utils import (
    raw_prediction_response_to_dataframe,
    recognize_sourcedata,
    get_id_from_response,
    from_api,
    retry,
)


def wait_for_async_predictions(project_id, predict_job_id, max_wait=600):
    """
    Given a Project id and PredictJob id poll for status of process
    responsible for predictions generation until it's finished

    Parameters
    ----------
    project_id : str
        The identifier of the project
    predict_job_id : str
        The identifier of the PredictJob
    max_wait : int, optional
        Time in seconds after which predictions creation is considered
        unsuccessful

    Returns
    -------
    predictions : pandas.DataFrame
        Generated predictions.

    Raises
    ------
    AsyncPredictionsGenerationError
        Raised if status of fetched PredictJob object is ``error``
    AsyncTimeoutError
        Predictions weren't generated in time, specified by ``max_wait``
        parameter
    """
    for index in retry.wait(max_wait):
        try:
            predict_job = PredictJob.get(project_id, predict_job_id)
        except errors.PendingJobFinished:
            return PredictJob.get_predictions(project_id, predict_job_id)
        if predict_job.status in (PREDICT_JOB_STATUS.ERROR, PREDICT_JOB_STATUS.ABORTED):
            e_msg = 'Predictions generation unsuccessful'
            raise errors.AsyncPredictionsGenerationError(e_msg)

    timeout_msg = 'Predictions generation timed out in {} seconds'.format(
        max_wait)
    raise errors.AsyncTimeoutError(timeout_msg)


class PredictJob(Job):
    _client = staticproperty(get_client)
    _fields = frozenset(['id',
                         'status',
                         'message'])
    _path = 'projects/{}/predictJobs/{}/'

    def __init__(self, data):
        self.id = None
        self.job_type = JOB_TYPE.PREDICT
        self.model = None
        self.project = None
        self.status = None
        self.message = None

        self._data = data
        if isinstance(data, dict):
            self._set_up(data)
        else:
            raise ValueError

    def _set_up(self, data):
        from . import Project, Model
        data = from_api(data)

        self.model = Model((data.get('project_id'), data.get('model_id')))
        self.project = Project(data.get('project_id'))

        converter = t.Dict({
            t.Key('id', optional=True) >> 'id': t.Int,
            t.Key('status', optional=True) >> 'status': t.String,
            t.Key('message', optional=True, default='') >> 'message': t.String(allow_blank=True),
        }).allow_extra('*')

        for k, v in six.iteritems(converter.check(data)):
            if k in self._fields:
                setattr(self, k, v)

        self._path = self._path.format(data.get('project_id'), data.get('id'))

    def __repr__(self):
        return 'PredictJob({}, status={})'.format(self.model,
                                                  self.status)  # pragma:no cover

    @classmethod
    def from_job(cls, job):
        """ Transforms a generic Job into a PredictJob

        Parameters
        ----------
        job: Job
            A generic job representing a PredictJob

        Returns
        -------
        predict_job: PredictJob
            A fully populated PredictJob with all the details of the job

        Raises
        ------
        ValueError:
            If the generic Job was not a predict job, e.g. job_type != JOB_TYPE.PREDICT
        """
        if not job.job_type == JOB_TYPE.PREDICT:
            raise ValueError('wrong job_type: {}'.format(job.job_type))
        data = job._get_full_details()
        return PredictJob(data)

    @classmethod
    def create(cls, model, sourcedata):
        """
        Starts predictions generation for provided data
        using previously created model.

        Parameters
        ----------
        model : Model
            Model to use for predictions generation
        sourcedata : str, file or pandas.DataFrame
            Data to be used for predictions.
            If str can be either a path to a local file or raw file content.

        Returns
        -------
        predict_job_id : str
            id of created job, can be used as parameter to ``PredictJob.get``
            or ``PredictJob.get_predictions`` methods or
            ``wait_for_async_predictions`` function

        Raises
        ------
        InputNotUnderstoodError
            If the parameter for sourcedata didn't resolve into known data types

        Examples
        --------
        .. code-block:: python

            model = Model.get('p-id', 'l-id')
            predict_job = PredictJob.create(model, './data_to_predict.csv')
        """
        url = 'projects/{}/predictions/'.format(model.project.id)
        form_data = {'model_id': model.id}
        default_fname = 'predict.csv'
        filesource_kwargs = recognize_sourcedata(sourcedata, default_fname)

        response = cls._client.build_request_with_file(
            url=url,
            form_data=form_data,
            method='post',
            **filesource_kwargs)
        return get_id_from_response(response)

    @classmethod
    def get(cls, project_id, predict_job_id):
        """
        Fetches one PredictJob. If the job finished, raises PendingJobFinished
        exception.

        Parameters
        ----------
        project_id : str
            The identifier of the project the model on which prediction
            was started belongs to
        predict_job_id : str
            The identifier of the predict_job

        Returns
        -------
        predict_job : PredictJob
            The pending PredictJob

        Raises
        ------
        PendingJobFinished
            If the job being queried already finished, and the server is
            re-routing to the finished predictions.
        AsyncFailureError
            Querying this resource gave a status code other than 200 or 303
        """
        url = cls._path.format(project_id, predict_job_id)
        response = cls._client.get(url, allow_redirects=False)
        if response.status_code == 200:
            data = response.json()
            return cls(data)
        elif response.status_code == 303:
            raise errors.PendingJobFinished
        else:
            e_msg = 'Server unexpectedly returned status code {}'
            raise errors.AsyncFailureError(
                e_msg.format(response.status_code))

    @classmethod
    def get_predictions(cls, project_id, predict_job_id):
        """
        Fetches finished predictions from the job used to generate them.

        Parameters
        ----------
        project_id : str
            The identifier of the project to which belongs the model used
            for predictions generation
        predict_job_id : str
            The identifier of the predict_job

        Returns
        -------
        predictions : pandas.DataFrame
            Generated predictions

        Raises
        ------
        JobNotFinished
            If the job has not finished yet
        AsyncFailureError
            Querying the predict_job in question gave a status code other than 200 or
            303
        """
        url = cls._path.format(project_id, predict_job_id)
        response = cls._client.get(url, allow_redirects=False)
        if response.status_code == 200:
            data = response.json()
            raise errors.JobNotFinished('Pending job status is {}'.format(data['status']))
        elif response.status_code == 303:
            location = response.headers['Location']
            response = cls._client.get(location, join_endpoint=False)
            return raw_prediction_response_to_dataframe(response.json())
        else:
            e_msg = 'Server unexpectedly returned status code {}'
            raise errors.AsyncFailureError(
                e_msg.format(response.status_code))

    def cancel(self):
        """
        Cancel this job. If this job has not finished running, it will be
        removed and canceled.
        """
        self._client.delete(self._path)



import trafaret as t
from ..client import get_client


class Feature(object):
    converter = t.Dict({
        'id': t.Int,
        'name': t.String,
        t.Key('featureType') >> 'feature_type': t.String,
        'importance': t.Or(t.Float, t.Null),
        t.Key('lowInformation') >> 'low_information': t.Bool,
        t.Key('uniqueCount') >> 'unique_count': t.Int,
        t.Key('naCount') >> 'na_count': t.Int,
    })

    def __init__(self, data):
        self.__dict__.update(self.converter.check(data))

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, self.name)

    @classmethod
    def get(cls, project_id, feature_id):
        """Retrieve a single feature

        Parameters
        ----------
        project_id : str
            The ID of the project the feature is associated with.
        feature_id : str
            The ID of the feature to retrieve

        Returns
        -------
        feature : Feature
            The queried instance
        """
        path = 'projects/{}/features/{}/'.format(project_id, feature_id)
        return Feature(get_client().get(path).json())



# flake8: noqa

__all__ = ['blueprint', 'job', 'modeljob', 'model', 'predict_job', 'project', 'featurelist']

from .model import Model
from .modeljob import ModelJob
from .blueprint import Blueprint
from .predict_job import PredictJob
from .featurelist import Featurelist
from .feature import Feature
from .job import Job
from .project import Project



import mock
import os
import six
import responses
import tempfile
from tempfile import NamedTemporaryFile
import unittest

import datarobot as dr
from datarobot.rest import RESTClientObject
from datarobot.client import (set_client, get_client, Client, _get_credentials_from_file,
                              _get_endpoint_from_file, _get_config_file)
from datarobot.errors import AppPlatformError
from .utils import SDKTestcase

if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch


class ClientTest(unittest.TestCase):

    def setUp(self):
        set_client(None)

    def tearDown(self):
        set_client(None)

    def test_instantiation(self):
        """
        Basic client installation.
        """
        with mock.patch(
            'os.environ',
                {'DATAROBOT_ENDPOINT': 'https://host_name.com'}):

            client = Client(token='t-token')
        assert get_client() is client
        for model in ('Featurelist', 'Model', 'ModelJob', 'Project', 'PredictJob'):
            model = getattr(dr, model)
            assert model._client is model({})._client is client

    @mock.patch('datarobot.client._file_exists', return_value=False)
    def test_instantiation_without_env(self, mock_file_args):
        """
        Basic client installation by get_client without configuration.
        """
        with mock.patch(
            'os.environ',
                {'DATAROBOT_ENDPOINT': 'https://host_name.com'}):
            with self.assertRaises(ValueError):
                get_client()

    def test_username_without_password_fails(self):
        with mock.patch(
            'os.environ',
                {'DATAROBOT_ENDPOINT': 'https://host_name.com',
                 'DATAROBOT_USERNAME': '',
                 'DATAROBOT_PASSWORD': ''}):
            with self.assertRaises(ValueError):
                Client(username='username')

    def test_password_without_username_fails(self):
        with mock.patch(
            'os.environ',
                {'DATAROBOT_ENDPOINT': 'https://host_name.com',
                 'DATAROBOT_USERNAME': '',
                 'DATAROBOT_PASSWORD': ''}):
            with self.assertRaises(ValueError):
                Client(password='password')

    def test_no_auth_fails(self):
        with mock.patch(
            'os.environ',
                {'DATAROBOT_ENDPOINT': 'https://host_name.com',
                 'DATAROBOT_USERNAME': '',
                 'DATAROBOT_PASSWORD': ''}):
            with self.assertRaises(ValueError):
                Client()

    @mock.patch('datarobot.client._file_exists')
    def test_no_endpoint_fails(self, mock_file_exists):
        mock_file_exists.return_value = False
        with mock.patch('os.environ', {}):
            with self.assertRaises(ValueError):
                Client(token='NOTATOKEN')

    def test_token_and_endpoint_is_okay(self):
        Client(token='token', endpoint='https://need_an_endpoint.com')

    def test_re_instantiation(self):
        """
        Client re installation.
        """
        with mock.patch('os.environ',
                        {'DATAROBOT_ENDPOINT': 'https://host_name.com'}):
            previous = Client(token='t-****')
            old_client = set_client(RESTClientObject(auth=('u-**********', 'p-******'),
                                                     endpoint='https://host_name.com'))
            self.assertIs(previous, old_client)

    @responses.activate
    def test_recognizing_domain_on_instance(self):
        raw = """{"api_token": "some_token"}"""
        responses.add(responses.POST,
                      'https://host_name.com/api/v2/api_token/',
                      body=raw,
                      status=201,
                      content_type='application/json')
        set_client(RESTClientObject(auth=('u-**********', 'p-******'),
                                    endpoint='https://host_name.com/api/v2'))
        restored_client = get_client()
        self.assertEqual(restored_client.domain, 'https://host_name.com')

    @mock.patch('datarobot.client._file_exists', return_value=False)
    def test_instantiation_from_env(self, mock_file_exists):
        """
        Test instantiation with creds from virtual environment
        """
        with patch.dict(
            'os.environ',
            {'DATAROBOT_API_TOKEN': 'venv_token',
             'DATAROBOT_ENDPOINT': 'https://host_name.com'}):
            rest_client = get_client()
            self.assertEqual(rest_client.token, 'venv_token')
            self.assertEqual(rest_client.endpoint, 'https://host_name.com')

        set_client(None)

        with patch.dict(
            'os.environ',
            {'DATAROBOT_API_TOKEN': 'venv_token',
             'DATAROBOT_ENDPOINT': 'https://host_name.com'}):
            rest_client = get_client()
            self.assertEqual(rest_client.token, 'venv_token')

    def test_instantiation_from_ini_file_with_env_path(self):
        raw_data = ("[datarobot]\n"
                    "token=file_token")
        with tempfile.NamedTemporaryFile() as test_file:
            test_file.write(str(raw_data).encode('utf-8'))
            test_file.seek(0)
            with patch('datarobot.client.os.environ',
                       {'DATAROBOT_CONFIG_FILE': test_file.name,
                        'DATAROBOT_ENDPOINT': 'https://host_name.com'}):
                rest_client = get_client()
        self.assertEqual(rest_client.token, 'file_token')

    def test_instantiation_from_yaml_file_with_env_path(self):
        raw_data = "token: file_token\n"
        with tempfile.NamedTemporaryFile(suffix=".yaml") as test_file:
            test_file.write(str(raw_data).encode('utf-8'))
            test_file.seek(0)
            with patch('datarobot.client.os.environ',
                       {'DATAROBOT_CONFIG_FILE': test_file.name,
                        'DATAROBOT_ENDPOINT': 'https://host_name.com'}):
                rest_client = get_client()
        self.assertEqual(rest_client.token, 'file_token')

    def test_instantiation_from_file_with_wrong_path(self):
        with patch.dict('os.environ', {'DATAROBOT_CONFIG_FILE': './tests/fixtures/.datarobotrc'}):
            with self.assertRaises(ValueError):
                get_client()

    def test_instantiation_from_ini_file_api_token(self):
        file_data = ('[datarobot]\n'
                     'token=fake_token\n'
                     'endpoint=https://host_name.com')
        config_file = tempfile.NamedTemporaryFile(delete=False)
        try:
            with open(config_file.name, mode='w') as config:
                config.write(file_data)
            fake_environ = {'DATAROBOT_CONFIG_FILE': config_file.name}
            with patch('os.environ', fake_environ):
                rest_client = get_client()
            self.assertEqual(rest_client.token, 'fake_token')
            self.assertEqual(rest_client.endpoint, 'https://host_name.com')
        finally:
            os.remove(config_file.name)

    def test_instantiation_from_yaml_file_api_token(self):
        file_data = ('token: fake_token\n'
                     'endpoint: https://host_name.com')
        config_file = tempfile.NamedTemporaryFile(delete=False, suffix='.yaml')
        try:
            with open(config_file.name, mode='w') as config:
                config.write(file_data)
            fake_environ = {'DATAROBOT_CONFIG_FILE': config_file.name}
            with patch('os.environ', fake_environ):
                rest_client = get_client()
            self.assertEqual(rest_client.token, 'fake_token')
            self.assertEqual(rest_client.endpoint, 'https://host_name.com')
        finally:
            os.remove(config_file.name)

    @mock.patch('datarobot.client._get_config_file')
    def test_instantiation_from_ini_file_default_path(self, mock_get_config_file):
        CONFIG_FILE = '.datarobotrc'
        mock_get_config_file.return_value = CONFIG_FILE
        raw_data = ("[datarobot]\n"
                    "token=file_token\n"
                    "endpoint=https://host_name.com")
        with mock.patch('os.environ', {}):
                if os.path.isfile(CONFIG_FILE):
                    # move existing config file out of test's way
                    os.rename(CONFIG_FILE, CONFIG_FILE + '.user')
                try:
                    with open(CONFIG_FILE, 'w+') as test_file:
                        test_file.write(raw_data)
                    self.assertTrue(test_file.closed)

                    rest_client = get_client()
                    self.assertEqual(rest_client.token, 'file_token')
                    self.assertEqual(rest_client.endpoint,
                                     'https://host_name.com')
                finally:
                    try:
                        os.remove(CONFIG_FILE)
                    except OSError:
                        pass
                    # may be bring user's config back
                    if os.path.isfile(CONFIG_FILE + ".user"):
                        os.rename(CONFIG_FILE + '.user', CONFIG_FILE)

    def test_client_from_codeline(self):
        Client(token='some_token',
               endpoint='https://endpoint.com')
        c = get_client()
        self.assertEqual(c.token, 'some_token')
        self.assertEqual(c.endpoint, 'https://endpoint.com')


class TestGetCredentials(unittest.TestCase):

    def setUp(self):
        os_patch = patch('datarobot.client.os')
        self.os_mock = os_patch.start()
        self.addCleanup(os_patch.stop)

        file_exists_patch = patch('datarobot.client._file_exists')
        self.exists_mock = file_exists_patch.start()
        self.addCleanup(file_exists_patch.stop)

        self.uname = 'user@domain.com'
        self.passwd = 'secrets'
        self.token = 'api-token'

    def test_token_ignored_when_uname_provided(self):
        self.os_mock.environ = dict(DATAROBOT_USERNAME=self.uname,
                                    DATAROBOT_PASSWORD=self.passwd,
                                    DATAROBOT_API_TOKEN=self.token)

        uname, passwd, token = dr.client.get_credentials_from_out()
        self.assertEqual(uname, self.uname)
        self.assertEqual(passwd, self.passwd)
        self.assertIsNone(token)

    def test_password_missing_from_env_results_in_none(self):
        self.os_mock.environ = dict(DATAROBOT_USERNAME=self.uname)
        uname, passwd, token = dr.client.get_credentials_from_out()
        self.assertEqual(uname, self.uname)
        self.assertIsNone(passwd)
        self.assertIsNone(token)

    def test_api_token_in_env_leaves_uname_passwd_blank(self):
        self.os_mock.environ = dict(DATAROBOT_API_TOKEN=self.token)
        uname, passwd, token = dr.client.get_credentials_from_out()
        self.assertIsNone(uname)
        self.assertIsNone(passwd)
        self.assertEqual(token, self.token)


class RestErrors(SDKTestcase):

    @responses.activate
    def test_404_plain_text(self):
        """
        Bad request in plain text
        """
        raw = "Not Found"

        responses.add(responses.GET, 'https://host_name.com/projects/404',
                      body=raw, status=404, content_type='text/plain')

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get('projects/404')

        self.assertEqual(str(app_error.exception),
                         '404 client error: Not Found')

    @responses.activate
    def test_404_json(self):
        """
        Bad request with datarobot reason in json
        """
        raw = """
        {"message": "Not Found"}
        """

        responses.add(responses.GET, 'https://host_name.com/projects/404',
                      body=raw, status=404, content_type='application/json')

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get('projects/404')

        exception_message = str(app_error.exception)
        self.assertIn('404 client error', exception_message)
        self.assertIn('Not Found', exception_message)

    @responses.activate
    def test_500_json(self):
        """
        Bad request with datarobot reason in json
        """
        raw = """
        {"message": "Not Found"}
        """

        responses.add(responses.GET, 'https://host_name.com/projects/500',
                      body=raw, status=500, content_type='application/json')

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get('projects/500')

        exception_message = str(app_error.exception)
        self.assertIn('500 server error', exception_message)
        self.assertIn('Not Found', exception_message)

    @responses.activate
    def test_other_errors(self):
        """
        Other errors
        """
        raw = """
        {"message": "Bad response"}
        """

        responses.add(responses.GET,
                      'https://host_name.com/projects/500',
                      body=raw, status=500, content_type='application/json')

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get('projects/500')
            self.assertEqual(str(app_error.exception),
                             'Connection refused: '
                             'https://host_name.com/projects/500')


class TestClientAttributes(unittest.TestCase):

    def test_main_useful_things_under_datarobot(self):
        """To lower the intimidation factor, let's try to limit the objects
        that show up at the root of datarobot

        This way, when they are in IPython and do tab-completion they get a
        sense for what is available to tinker with at the top level
        """
        known_names = {'Project',
                       'Model',
                       'Blueprint',
                       'ModelJob',
                       'PredictJob',
                       'Job',
                       'QUEUE_STATUS',
                       'Client',
                       'AUTOPILOT_MODE',
                       'AppPlatformError',
                       'utils',
                       'errors',
                       'models',
                       'client',
                       'rest',
                       'SCORING_TYPE',
                       'Featurelist',
                       'Feature',
                       'helpers',
                       'RandomCV',
                       'StratifiedCV',
                       'GroupCV',
                       'UserCV',
                       'RandomTVH',
                       'UserTVH',
                       'StratifiedTVH',
                       'GroupTVH',
                       'partitioning_methods',
                       'RecommenderSettings',  # TODO: Too many attrs
                       'AdvancedOptions',
                       'VERBOSITY_LEVEL',
                       'enums'
                       }

        found_names = {name for name in dir(dr)
                       if not (name.startswith('__') and name.endswith('__'))}
        assert found_names == known_names


class TestProcessConfigFile(unittest.TestCase):

    @patch('datarobot.client._file_exists')
    @patch('datarobot.client.os', autospec=True)
    def test_get_config_file_from_environ(self, mock_os, mock_file_exists):
        mock_os.environ = {'DATAROBOT_CONFIG_FILE': 'fake_config_file'}
        mock_file_exists.return_value = True
        config_file = _get_config_file()
        self.assertEqual(config_file, 'fake_config_file')

    @patch('datarobot.client._file_exists')
    @patch('datarobot.client.os', autospec=True)
    def test_get_config_file_from_environ_dne(self, mock_os, mock_file_exists):
        mock_os.environ = {'DATAROBOT_CONFIG_FILE': 'fake_config_file'}
        mock_file_exists.return_value = False
        with self.assertRaises(ValueError):
            _get_config_file()

    @patch('datarobot.client._file_exists')
    @patch('datarobot.client.os.environ')
    def test_get_config_file_deprecated_exists(self, mock_os_environ, mock_file_exists):
        mock_os_environ.data = {}
        deprecated_config_path = os.path.join(os.path.expanduser('~'), '.datarobotrc')
        mock_file_exists.side_effect = \
            lambda path: True if path == deprecated_config_path else False
        self.assertEqual(_get_config_file(), deprecated_config_path)

    @patch('datarobot.client._file_exists')
    @patch('datarobot.client.os.environ')
    def test_get_config_file_both_exist(self, mock_os_environ, mock_file_exists):
        first_choice_config_path = os.path.expanduser('~/.config/datarobot/drconfig.yaml')
        mock_os_environ.data = {}
        mock_file_exists.return_value = True
        self.assertEqual(_get_config_file(), first_choice_config_path)

    @patch('datarobot.client._file_exists')
    @patch('datarobot.client.os', autospec=True)
    def test_get_default_config_file_dne(self, mock_os, mock_file_exists):
        mock_os.environ = {}
        mock_file_exists.return_value = False
        mock_os.path.join.return_value = '/i/dont/exist'
        self.assertIsNone(_get_config_file())

    def test_get_credentials_from_ini_file_present(self):
        file_content = ('[datarobot]\n'
                        'username=user@email.domain\n'
                        'password=file_password\n'
                        'token=fake_token')
        fake_config_file = NamedTemporaryFile(delete=False)
        try:
            with open(fake_config_file.name, mode='w') as config_file:
                config_file.write(file_content)
            username, password, token = _get_credentials_from_file(fake_config_file.name)
            self.assertEqual(username, 'user@email.domain')
            self.assertEqual(password, 'file_password')
            self.assertEqual(token, 'fake_token')
        finally:
            os.remove(fake_config_file.name)

    def test_get_credentials_from_yaml_file_present(self):
        file_content = ('username: user@email.domain\n'
                        'password: file_password\n'
                        'token: fake_token')
        fake_config_file = NamedTemporaryFile(delete=False, suffix='.yaml')
        try:
            with open(fake_config_file.name, mode='w') as config_file:
                config_file.write(file_content)
            username, password, token = _get_credentials_from_file(fake_config_file.name)
            self.assertEqual(username, 'user@email.domain')
            self.assertEqual(password, 'file_password')
            self.assertEqual(token, 'fake_token')
        finally:
            os.remove(fake_config_file.name)

    def test_get_credentials_from_file_file_dne(self):
        fake_config_file = NamedTemporaryFile()
        fake_config_file.close()
        self.assertFalse(os.path.exists(fake_config_file.name))
        username, password, token = _get_credentials_from_file(fake_config_file.name)
        self.assertIsNone(username)
        self.assertIsNone(password)
        self.assertIsNone(token)

    def test_get_credentials_from_file_partial_cred(self):
        file_content = ('[datarobot]\n'
                        'token=fake_token')
        fake_config_file = NamedTemporaryFile(delete=False)
        try:
            with open(fake_config_file.name, mode='w') as config_file:
                config_file.write(file_content)
            username, password, token = _get_credentials_from_file(fake_config_file.name)
            self.assertIsNone(username)
            self.assertIsNone(password)
            self.assertEqual(token, 'fake_token')
        finally:
            os.remove(fake_config_file.name)

    def test_get_endpoint_from_ini_file_succeess(self):
        file_content = ('[datarobot]\n'
                        'endpoint=http://host_name.com')
        fake_config_file = NamedTemporaryFile(delete=False)
        try:
            with open(fake_config_file.name, mode='w') as config_file:
                config_file.write(file_content)
        finally:
            os.remove(fake_config_file.name)

    def test_get_endpoint_from_yaml_file_succeess(self):
        file_content = ('endpoint: http://host_name.com')
        fake_config_file = NamedTemporaryFile(delete=False, suffix='.yaml')
        try:
            with open(fake_config_file.name, mode='w') as config_file:
                config_file.write(file_content)
        finally:
            os.remove(fake_config_file.name)

    def test_get_endpoint_from_file_file_dne(self):
        fake_config_file = NamedTemporaryFile()
        fake_config_file.close()
        self.assertFalse(os.path.exists(fake_config_file.name))
        self.assertIsNone(_get_endpoint_from_file(fake_config_file.name))

    def test_get_endpoint_from_file_missing(self):
        file_content = ('[datarobot]\n'
                        'token=fake_token')
        fake_config_file = NamedTemporaryFile(delete=False)
        try:
            with open(fake_config_file.name, mode='w') as config_file:
                config_file.write(file_content)
            self.assertIsNone(_get_endpoint_from_file(fake_config_file.name))
        finally:
            os.remove(fake_config_file.name)



import six
import trafaret as t
from ..client import get_client, staticproperty
from ..utils import from_api


class Featurelist(object):
    _client = staticproperty(get_client)
    _fields = frozenset(['id',
                         'name',
                         'features'])
    _path = 'projects/{}/featurelists/'

    def __repr__(self):
        return 'Featurelist({})'.format(self.name)

    def __init__(self, data):
        self.id = None
        self.name = None
        self.features = None
        self.project = None
        self._data = data

        if isinstance(data, dict):
            self._set_up(data)
        else:
            raise ValueError

    def _set_up(self, data):
        from . import Project

        data = from_api(data)

        self.project = Project(data.get('project_id'))

        converter = t.Dict({
            t.Key('id', optional=True) >> 'id': t.String(allow_blank=True),
            t.Key('name', optional=True) >> 'name': t.String,
            t.Key('features', optional=True) >> 'features': t.List(t.String),
        }).allow_extra("*")

        for k, v in six.iteritems(converter.check(data)):
            if k in self._fields:
                setattr(self, k, v)

        self._path = self._path.format(data.get('project_id'))

    @classmethod
    def get(cls, project, featurelist_id):
        """Retrieve a known feature list

        Parameters
        ----------
        project : Project instance or str
            The project the featurelist is associated with, specified either
            with an instance of Project or the project id
        featurelist_id : str
            The ID of the featurelist to retrieve

        Returns
        -------
        featurelist : Featurelist
            The queried instance
        """
        from . import Project

        if isinstance(project, Project):
            project_id = project.id
        elif isinstance(project, six.string_types):
            project_id = project
        else:
            raise ValueError('Project arg must be Project instance or str')

        url = cls._path.format(project_id) + featurelist_id + '/'
        resp_data = cls._client.get(url).json()
        return cls(resp_data)



import warnings


class AppPlatformError(Exception):
    """
    Raised by :meth:`Client.request()` for requests that:
      - Return a non-200 HTTP response, or
      - Connection refused/timeout or
      - Response timeout or
      - Malformed request
      - Have a malformed/missing header in the response.
    """
    def __init__(self, exc_message, status_code, error_code=None):
        super(AppPlatformError, self).__init__(exc_message)
        self.status_code = status_code
        self.error_code = error_code


class ServerError(AppPlatformError):
    """
    For 500-level responses from the server
    """


class ClientError(AppPlatformError):
    """
    For 400-level responses from the server
    """


class InputNotUnderstoodError(Exception):
    """
    Raised if a method is called in a way that cannot be understood
    """


class AllRetriesFailedError(Exception):
    """Raised when the retry manager does not successfully make a request"""


class InvalidModelCategoryError(Exception):
    """
    Raised when method specific for model category was called from wrong model
    """


class AsyncTimeoutError(Exception):
    """
    Raised when an asynchronous operation did not successfully get resolved
    within a specified time limit
    """


class AsyncFailureError(Exception):
    """
    Raised when querying an asynchronous status resulted in an exceptional
    status code (not 200 and not 303)
    """


class ProjectAsyncFailureError(AsyncFailureError):
    """
    When an AsyncFailureError occurs during project creation or finalizing the project
    settings for modeling. This exception will have the attributes ``status_code``
    indicating the unexpected status code from the server, and ``async_location`` indicating
    which asynchronous status object was being polled when the failure happened.
    """
    def __init__(self, exc_message, status_code, async_location):
        super(ProjectAsyncFailureError, self).__init__(exc_message)
        self.status_code = status_code
        self.async_location = async_location


class AsyncProcessUnsuccessfulError(Exception):
    """
    Raised when querying an asynchronous status showed that async process
    was not successful
    """


class AsyncModelCreationError(Exception):
    """
    Raised when querying an asynchronous status showed that model creation
    was not successful
    """


class AsyncPredictionsGenerationError(Exception):
    """
    Raised when querying an asynchronous status showed that predictions
    generation was not successful
    """


class PendingJobFinished(Exception):
    """
    Raised when the server responds with a 303 for the pending creation of a
    resource.
    """


class JobNotFinished(Exception):
    """
    Raised when execution was trying to get a finished resource from a pending
    job, but the job is not finished
    """


class DuplicateFeaturesError(Exception):
    """
    Raised when trying to create featurelist with duplicating features
    """


class DataRobotDeprecationWarning(DeprecationWarning):
    """
    Raised when using deprecated functions or using functions in a deprecated way
    """
    pass


warnings.filterwarnings('default', category=DataRobotDeprecationWarning)



import json
import responses
from datarobot import Feature, Project

data = {
    'id': 34,
    'name': 'Claim_Amount',
    'featureType': 'Numeric',
    'importance': 1,
    'lowInformation': False,
    'uniqueCount': 200,
    'naCount': 0,
}


@responses.activate
def test_features(client):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/',
                  body=json.dumps([data]))
    feature, = Project({'id': 'p-id'}).get_features()
    assert isinstance(feature, Feature)
    assert feature.id == 34
    assert feature.name == 'Claim_Amount'
    assert feature.feature_type == 'Numeric'
    assert feature.importance == 1
    assert feature.low_information is False
    assert feature.unique_count == 200
    assert feature.na_count == 0


@responses.activate
def test_feature(client):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/f-id/',
                  body=json.dumps(dict(data, importance=None)))
    feature = Feature.get('p-id', 'f-id')
    assert isinstance(feature, Feature)
    assert feature.id == 34
    assert feature.name == 'Claim_Amount'
    assert feature.feature_type == 'Numeric'
    assert feature.importance is None
    assert feature.low_information is False
    assert feature.unique_count == 200
    assert feature.na_count == 0



import mock
import six

from datarobot.utils import retry
zip = six.moves.zip


def test_wait():
    delay, maxdelay = retry.wait.__defaults__
    assert list(index for index, _ in retry.wait(0)) == [0]
    assert list(index for index, _ in retry.wait(delay)) == [0, 1]
    it = (index for index, _ in retry.wait(float('inf')))
    expected = 0.1, 0.2, 0.4
    with mock.patch('datarobot.utils.retry.time.sleep') as sleep:
        assert next(it) == 0 and not sleep.called
        for index, delay in zip(it, expected):
            assert sleep.call_args[0] == (delay,)
    assert index == len(expected)



"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

import platform
import os

import requests
import six
from requests_toolbelt.multipart.encoder import MultipartEncoder
from six.moves.urllib_parse import urlparse

from . import __version__
from .utils import to_api, get_endpoint
from . import errors


class RESTClientObject(requests.Session):
    """
    Parameters
        default_timeout
            timeout for http request and connection
        headers
            headers for outgoing requests
    """

    default_timeout = 60

    def __init__(self, auth=None, endpoint=None):
        super(RESTClientObject, self).__init__()
        # Note: As of 2.3, get_endpoint will be gone and we will make `endpoint` required
        self.endpoint = endpoint or get_endpoint()
        self.domain = '{}://{}'.format(urlparse(self.endpoint).scheme,
                                       urlparse(self.endpoint).netloc)
        self.token = auth
        self.user_agent_header = self._make_user_agent_header()
        self.headers.update(self.user_agent_header)
        self.token_header = {'Authorization': 'Token {}'.format(self.token)}
        self.headers.update(self.token_header)

    def _make_user_agent_header(self):
        client_str = 'DataRobotPythonClient/{}'.format(__version__)
        platform_str = '{} {} {}'.format(platform.system(), platform.release(), platform.machine())
        final_header = {'User-Agent': '{} ({})'.format(client_str, platform_str)}
        return final_header

    def _join_endpoint(self, url):
        return '{}/{}'.format(self.endpoint, url)

    def strip_endpoint(self, url):
        expected = '{}/'.format(self.endpoint)
        if not url.startswith(expected):
            raise ValueError('unexpected url format: {}'.format(url))
        return url.split(expected)[1]

    def request(self, method, url, join_endpoint=True, **kwargs):
        kwargs.setdefault('timeout', self.default_timeout)
        if join_endpoint:
            url = self._join_endpoint(url)
        response = super(RESTClientObject, self).request(method, url, **kwargs)
        if not response:
            handle_http_error(response)
        return response

    def get(self, url, params=None, **kwargs):
        return self.request('get', url, params=to_api(params), **kwargs)

    def post(self, url, data=None, **kwargs):
        if data:
            kwargs['json'] = to_api(data)
        return self.request('post', url, **kwargs)

    def patch(self, url, data=None, **kwargs):
        if data:
            kwargs['json'] = to_api(data)
        return self.request('patch', url, **kwargs)

    def build_request_with_file(self, method, url,
                                fname,
                                form_data=None,
                                content=None,
                                file_path=None,
                                filelike=None):
        """Build request with a file that will use special
        MultipartEncoder instance (lazy load file).


        This method supports uploading a file on local disk, string content,
        or a file-like descriptor. ``fname`` is a required parameter, and
        only one of the other three parameters can be provided.

        Parameters
        ----------
        method : str.
            Method of request. This parameter is required, it can be
            'POST' or 'PUT' either 'PATCH'.
        url : str.
            Url that will be used it this request.
        fname : name of file
            This parameter is required, even when providing a file-like object
            or string content.
        content : str
            The content buffer of the file you would like to upload.
        file_path : str
            The path to a file on a local file system.
        filelike : file-like
            An open file descriptor to a file.

        Returns
        -------
        response : response object.

        """

        bad_args_msg = ('Upload should be used either with content buffer '
                        'or with path to file on local filesystem or with '
                        'open file descriptor')
        assert sum((bool(content),
                    bool(file_path),
                    bool(filelike))) == 1, bad_args_msg

        if file_path:
            if not os.path.exists(file_path):
                raise ValueError(
                    'Provided file does not exist {}'.format(file_path))
            fields = {'file': (fname, open(file_path, 'rb'))}

        elif filelike:
            filelike.seek(0)
            fields = {'file': (fname, filelike)}
        else:
            if not isinstance(content, six.binary_type):
                raise AssertionError('bytes type required in content')
            fields = {'file': (fname, content)}

        form_data = form_data or {}
        data_for_encoder = to_api(form_data)
        data_for_encoder.update(fields)

        encoder = MultipartEncoder(fields=data_for_encoder)
        headers = {'Content-Type': encoder.content_type}
        return self.request(method, url, headers=headers, data=encoder)


def _http_message(response):
    if response.status_code == 401:
        message = ('The server is saying you are not properly '
                   'authenticated. Please make sure your API '
                   'token is valid.')
    elif response.status_code == 403:
        message = (
            'The server is saying that you do not have '
            'permissions to do the operation you '
            'requested, even though your API token is '
            'valid.')
    elif response.headers['content-type'] == 'application/json':
        message = response.json()
    else:
        message = response.content.decode('ascii')[:79]
    return message


def handle_http_error(response):
    message = _http_message(response)
    if 400 <= response.status_code < 500:
        template = '{} client error: {}'
        exc_message = template.format(response.status_code, message)
        raise errors.ClientError(exc_message, response.status_code)
    else:
        template = '{} server error: {}'
        exc_message = template.format(response.status_code, message)
        raise errors.ServerError(exc_message, response.status_code)



import json

import responses

from datarobot import Job, Project


mock_job_data = {
    'status': 'inprogress',
    'url': 'https://host_name.com/projects/p-id/modelJobs/1/',
    'id': '1',
    'jobType': 'model',
    'projectId': 'p-id'
}


mock_model_job_data = {
    'status': 'inprogress',
    'processes': [
        'One-Hot Encoding',
        'Bernoulli Naive Bayes classifier (scikit-learn)',
        'Missing Values Imputed',
        'Gaussian Naive Bayes classifier (scikit-learn)',
        'Naive Bayes combiner classifier',
        'Calibrate predictions'
    ],
    'projectId': 'p-id',
    'samplePct': 28.349,
    'modelType': 'Naive Bayes combiner classifier',
    'featurelistId': '56d8620bccf94e26f37af0a3',
    'modelCategory': 'model',
    'blueprintId': '2a1b9ae97fe61880332e196c770c1f9f',
    'id': '1'
}


def test_instantiate(client):
    new_job = Job(mock_job_data)
    assert new_job.id == 1
    assert new_job.project == Project('p-id')
    assert new_job.status == 'inprogress'
    assert new_job.job_type == 'model'
    assert new_job._path == 'projects/p-id/modelJobs/1/'


@responses.activate
def test_get_full_details(client):
    body = json.dumps(mock_model_job_data)
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/modelJobs/1/',
        status=200,
        body=body,
        content_type='application/json'
    )
    generic_job = Job(mock_job_data)
    model_job_dict = generic_job._get_full_details()
    assert model_job_dict == mock_model_job_data


@responses.activate
def test_cancel_job(client):
    responses.add(
        responses.DELETE,
        'https://host_name.com/projects/p-id/modelJobs/1/',
        status=204,
        body='',
        content_type='application/json'
    )
    generic_job = Job(mock_job_data)
    generic_job.cancel()
